/*
 * $Id: knapsack.cilk,v 1.2 1996/06/04 16:32:25 athena Exp $
 */

/*
 * Cilk program to solve the 0-1 knapsack problem using a branch-and-bound
 * technique. Test for aborts.
 *
 * Adapted for REAPAR by Stefan Haenssgen:
 *
 * 06-jan-97	Removed cilk-specific functions (cilk, sync, spawn)
 *		Converted knapsack() to void, added reference parameter
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

/* every item in the knapsack has a weight and a value */
#define MAX_ITEMS 256

struct item {
     int value;
     int weight;
};

struct item items[MAX_ITEMS];	/* array of items */

int capacity;
int nitems;
int best_so_far = INT_MIN;

int compare(struct item *a, struct item *b)
{
     double c = ((double) a->value / a->weight) -
     ((double) b->value / b->weight);

     if (c > 0)
	  return -1;
     if (c < 0)
	  return 1;
     return 0;
}

void read_input(void)
{
     int i;

     /* format of the input: #items capacity value1 weight1 ... */
     scanf("%d", &nitems);
     scanf("%d", &capacity);

     for (i = 0; i < nitems; ++i)
	  scanf("%d %d", &items[i].value, &items[i].weight);

     /* sort the items on decreasing order of value/weight */
     /* cilk2c is fascist in treting pointers, whence the ugly cast */
     qsort(items, nitems, sizeof(struct item),
	    (int (*)(const void *, const void *)) compare);
}

/* 
 * return the optimal solution for n items (first is e) and
 * capacity c. Value so far is v.
 */
void knapsack(int *retval, struct item *e, int c, int n, int v)
{
     int with, without, best;
     double ub;

     /* base case: full knapsack or no items */
     if (c <= 0) {
	  *retval = INT_MIN;	/* infeasible solution */
	  return;
     }

     if (n == 0) {
	  *retval = v;		/* feasible solution, with value v */
	  return;
     }

     ub = (double) v + c * e->value / e->weight;

     if (ub < best_so_far) {
	  /* prune ! */
	  *retval = INT_MIN;
	  return;
     }
     
     /* compute the best solution with the current item in the knapsack
      */
     knapsack(&with, e + 1, c - e->weight, n - 1, v + e->value);

     /* compute the best solution without the current item in the knapsack 
      */
     knapsack(&without, e + 1, c, n - 1, v);	/* NOTHREAD */

     /* NEEDRESULTS */
     
     if (with > without)
	  best = with;
     else
	  best = without;

     /* 
      * notice the right condition here. The program is still
      * correct, in the sense that the best solution so far
      * is at least best_so_far. Moreover best_so_far gets updated
      * when returning, so eventually it should get the right
      * value. The program is highly non-deterministic.
      */
     if (best > best_so_far)
	  best_so_far = best;

     *retval = best;
}

int main()
{
     int sol;

     read_input();
     knapsack(&sol, items, capacity, nitems, 0);

     printf("Best value is %d\n", sol);

     return 0;
}
