/*
 * simulation.c
 *
 * Parses tree input from statistics generation and simulates
 * keep_n/active_n/depth_n thread generation strategies
 *
 *
 * Usage:   simulation [options] < treefile
 *	    Options:
 *	    -s A	Simulate always strategy (also '-A')
 *	    -s a [n]	Simulate active_n strategy (also '-a [n]')
 *	    -s d [n]	Simulate depth_n strategy (also '-d [n]')
 *	    -s f [n]	Simulate first_n strategy (also '-f [n]')
 *	    -s k [n]	Simulate keep_n strategy (also '-k [n]', default)
 *	    -s n	Simulate never strategy (also '-n')
 *	    -c C	Simulate C CPUs (default = 4)
 *	    -p N	Strategy Parameter: N*C threads with keep/active
 *			 resp. depth N (default = 3)
 *	    -t ABCDE..	Only execute procedure type A/B/... in parallel
 *			(A=1st recursive procedure etc, default=all)
 *	    -b Ax	Don't create thread for branch #x of proc. A
 *			(A as above, branches count starting from 0)
 *	    -e		Execute non-parallel parts in one single step
 *	    -v		Verbose output
 *	    -q		Quiet output, don't print final tree with threads
 *	    treefile	ASCII representation of recursion tree, e.g.
 *			"(()())" for simple binary tree, may be annotated
 *			with procedure numbers and subtree timings
 *         Output format for one node: PDTTTS
 *             P   = Node's procedure (A,B,...)
 *             D   = Node's state ('.' incomplete / 'd' done)
 *             TTT = Thread ID working on that node (000 = none)
 *             S   = Thread's status (None, Active, Waiting, Done, Joined)
 *
 *
 * Stefan Haenssgen  04-oct-97
 *
 *
 * 04-oct-97	1st version based on tree_parser.c
 *		- Mark nodes with thread ID
 *		- Different thread and node states
 *		- Include Link to parent node
 *		- State machine for thread behavior
 *		- Work loop for thread until state changes
 * 06-oct-97	Also take CPU number into account, 1 thread per CPU
 * 08-oct-97	More detailed output in verbose setting
 *		Debugged switching/on_cpu setting
 *		Also record number of nodes/leafs done per thread
 *		Better check for children-done in active state
 *		Yield CPU after new thread creation
 *		Added thread waiting queue
 * 09-oct-97	Also record thread's starting and maximum recursion depth
 *		Save depth in nodes for easier reference
 * 10-oct-97	Start with thread 0 (original process) instead of 1,
 *		 adapted all checks accordingly (flag '-1' instead of '0'
 *		 for empty/not used thread)
 *		Change (based on observations of real Solaris behavior):
 *		 Do NOT yield cpu after creating a new thread, continue
 *		 computation!
 * 14-oct-97	HACK: only check nodes whose type is the same as
 *		 the root node's type (later, we'll have to add real
 *		 procedure type handling!)
 * 15-oct-97	Real procedure type check: Parameter '-t ABC..' to
 *		 activate procedure 1/2/3 (default = all) -> parprocs[]
 *		Obey parprocs, i.e. skip over nodes whose procedure
 *		 type is not marked as parallel in parprocs
 *		Added skipping mode for that
 *		As part of node initialization, mark node as "parleaf"
 *		 if all its subnodes have only non-parallel proc-types
 *		Stop skipping mode if node is done
 *		Added proc struct with par-annotation and branching
 *		 degree (nothread) annotation
 *		Obey this structure, i.e. don't create thread for that
 *		 procedure at the given branching degree
 *		New termination check: root node done, not 1st thread done
 *		 as before
 * 16-oct-97	Strategies selected with "-s" option
 *		Give help output on any unknown parameter
 *		Added DEPTH_N strategy
 * 17-oct-97	Measure CPU's load over time (simsteps, cpusbusy) - a
 *		 CPU is busy if it has a thread and the thread doesn't
 *		 just sit around and wait for children
 *		Also count nodes/leafs of non-parallel procedures
 *		 a thread encounters (in subnodes/subleafs)
 * 23-oct-97	Process switch emulation - hm, generates more threads but
 *		 not very useful right now...
 *		Strategy recommendations based on heuristics from depth
 *		 strategy, but now refering to threads insteas of subtrees
 *		Corresponding statistics for (sub)nodes/leafs
 * 27-oct-97	Statistics for procedures (number of [par]nodes/leafs etc)
 *		Also recognize "-a 5" as active 5, i.e. "-p" not necessary
 *		Optionally exectute non-parallel parts in single
 *		 steps, too (option "-e", "seqstep" flag)
 *		Reversed the "-e" logic - default = single steps, "seqonestep"
 *		Also consider simulated load in heuristics, global allcpuload
 * 04-dec-97	Switch "-q" for quiet mode, no final tree printing
 *		Added FIRST_N, ALWAYS and NEVER strategies (-f, -A, -n)
 * 16-jan-98	Exit simulation with warning if number of threads in queue
 *		 exceeds 3000 (that's when Solaris aborts in reality)
 * 22-jan-98	DEPTH strategy says "current depth < D", not "<=" ! Fixed
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define TEXTCHUNKSIZE 16384 /* Size of text string chuncs (4 allocation)*/

#define CHILDNUM 10	/* Default starting number of children for nodes*/

typedef struct _node *nodeptr;

typedef struct _node {
  int      numchildren;	/* Number of children this node has		*/
  nodeptr  parent;	/* Parent node					*/
  nodeptr *children;	/* Array of children for this node		*/
  int 	   numnodes;	/* Number of nodes (without leafs) in subtree	*/
  int      numleafs;	/* Number of leafs in this subtree		*/
  int      procnum;	/* Number code for the node's procedure		*/
  long     cputime;	/* Used CPU time for this subtree in millisec.	*/
  int      threadid;	/* ID of thread working on this node (0=none)	*/
  char     nodedone;    /* 1 if node has been worked on (including subs)*/
  char	   parleaf;	/* 1 if all subtrees have non-parallel procs	*/
  int	   depth;	/* Depth of this node in recursion tree		*/
  int     *info;	/* Any additional information (size, ...)	*/
} node;

#define NULLPTR (nodeptr)NULL;


/* Thread states */

#define T_NONE    0	/* Not initialized/active yet			*/
#define T_ACTIVE  1	/* Working on the current node			*/
#define T_WAITING 2	/* Waiting for children to finish		*/
#define T_DONE    3	/* No more work to do				*/
#define T_JOINED  4	/* Finished and joined by parent thread		*/

#define statechar(s) ((s==T_NONE)?'N':(				\
                          (s==T_ACTIVE)?'A':(			\
			      (s==T_WAITING)?'W':(		\
				  (s==T_DONE)?'D':(		\
				      (s==T_JOINED)?'J':'?'	\
		      )))))
                        
typedef struct {
    int  state;		/* One of the above states			*/
    char on_cpu;	/* 1 if a CPU is working on this thread		*/
    long nodesdone;	/* Number of nodes visited (worked on)		*/
    long leafsdone;	/* Ditto leafs (only of parallel procedures!)	*/
    long subnodesdone;	/* Number of nodes/leafs visited (worked on)	*/
    long subleafsdone;	/*  but also including non-parallel procedures	*/
    int  startdepth;	/* Starting recursion depth in tree		*/
    int  maxdepth;	/* Maximum depth reached			*/
    node *startnode;	/* Node from which this thread was started	*/
    node *currentnode;	/* Node this thread is currently working on	*/
} thread;

#define MAXTHREADS 1000000	/* Maximum number of threads simulated	*/

thread threads[MAXTHREADS];	/* Thread information array		*/

int threadq[MAXTHREADS];	/* Queue of threads ready to run 	*/
int threadqlen;			/* (excluding done/joined and cpu-bound)*/


/* CPUs */

#define MAXCPUS 1024		/* Maximum number of CPUs simulated	*/

int    cputhreads[MAXCPUS];	/* Current thread for each CPU (0=idle)	*/

int   *cpusbusy;		/* Number of busy CPUs in one sim step	*/
int    simsteps;		/* Number of simulation steps		*/
int    simstepsalloc;		/* Number of currently allocated steps	*/
double allcpuload;	/* CPU load percentage (max. CPUnum * 100)	*/

char *text;		/* Global pointer to current input position	*/

int maxprocnum;		/* Global maximum number of proc-indices found	*/
                        /* Zero here means no proc-numbers were found   */
                        /* at all. procnums range 1..mp, arrays 0..mp-1	*/

#define S_KEEP    1	/* KEEP_N Strategy				*/
#define S_ACTIVE  2	/* ACTIVE_N Strategy				*/
#define S_DEPTH   4	/* DEPTH_N Strategy				*/
#define S_FIRST   8	/* FIRST_N Strategy				*/
#define S_ALWAYS 16	/* ALWAYS Strategy				*/
#define S_NEVER  32	/* NEVER Strategy				*/

int cpunum;		/* Number of simulated CPUs			*/
int stratparam;		/* Parameter: Number of threads per CPU/Depth	*/
int strategy;		/* Parallelization strategy (KEEP or ACTIVE)	*/

int numthreads;		/* Global number of threads generated yet	*/
int numactive;		/* Ditto currently active (depends on strategy)	*/


/* Procedures */

#define MAXPROCS 128	/* Maximum number of recursive procedures	*/
#define MAXDEGREE 32	/* Maximum number of branches in one node	*/

typedef struct {
  char	par;			/* 1 if procedure parallelized, 0 else	*/
  char  nothread[MAXDEGREE];	/* 1 if no therad annotated for degree	*/
  long  nodes;			/* Number of nodes/leafs from this proc.*/
  long  leafs;
  long  parnodes;		/* Ditto for parallel "leafs"/nodes	*/
  long  parleafs;
} proc;

proc procs[MAXPROCS];	/* Global array of procedure information	*/


char seqonestep = 0;	/* 1 if sequential procs executed as one big step */

char quietmode = 0;	/* 1 if final tree with threads not to be printed */

/*long procswitch = 0;	/* for global process switch emulation	*/


/* Forward declarations */

nodeptr parse_tree();
void    print_tree(nodeptr n, int depth, char first,
		   char verbose, char procverbose);
int     tree_depth(nodeptr n);
void    print_subtrees_at_depth(nodeptr n, int depth,
				char verbose, char procverbose);
nodeptr subtrees_at_depth(nodeptr n, int depth);
void    get_subtrees_at_depth(nodeptr n, int depth, nodeptr statnode);
void    print_tree_graphics(nodeptr n, int depth,
	   		    char first, char last, char* branches,
			    char verbose);


/* For quicksort
 */
static int intcompare(int *i, int *j)
{
    if (*i > *j)
	return (1);
    if (*i < *j)
	return (-1);
    return (0);
}


/* Sorts array of integers using library quicksort
 */
void intsort(int *a, int len)
{
  qsort((char *) a, len, sizeof(int), intcompare);
}


/* Parse global input text recursively:
 * "(" opens a new node and descends into it, ")" returns from a node
 */
nodeptr parse_tree() {
  char     c;			/* Current character in input	*/
  int      count;		/* Count children		*/
  nodeptr  n;			/* This node			*/
  nodeptr *cs;			/* Child pointer for this node	*/
  int      maxchildren;		/* Currently allocated #children*/
  int      procnum = 0;		/* Procedure code-number 	*/
  char	   haveprocnum = 0;	/* True procnum	already read	*/
  long     cputime = 0;		/* ms CPU Time spent in subtree	*/


  n = (nodeptr) malloc(sizeof(node));
  n->numchildren = 0;		/* Init values			*/
  n->parent   = NULLPTR;
  n->children = (nodeptr*)0;
  n->numleafs = 0;
  n->numnodes = 0; 
  n->procnum  = 0;
  n->cputime  = 0;
  n->threadid = -1;		/* No thread visited node yet	*/
  n->nodedone = 0;		/* Node not yet done		*/
  n->info = NULL;
  

  while (c = *(text++)) {	/* Read next character		*/

    if (c == '(') {		/* New child, descend		*/

      (n->numchildren)++;
      if (n->children == 0) {
	
	/* Didn't have children yet, allocate some space	*/

	n->children = (nodeptr*)malloc(CHILDNUM * sizeof(nodeptr));
	maxchildren = CHILDNUM;	/* Remember current max.alloc.	*/

      } else if (n->numchildren >= maxchildren) {
	
	/* More children than already allocated, make room	*/

	maxchildren = maxchildren + CHILDNUM;
	n->children = (nodeptr*)realloc( (void*) n->children,
					 maxchildren * sizeof(nodeptr));
      }

      n->children[n->numchildren-1] = parse_tree();
      n->children[n->numchildren-1]->parent = n;/* Set parent	*/
      n->numleafs = n->numleafs +		/* Count leafs	*/
	            n->children[n->numchildren-1]->numleafs;
      n->numnodes = n->numnodes +		/* Count nodes	*/
	            n->children[n->numchildren-1]->numnodes;

    } else if (c == ')') {	/* End of node, null		*/
      
      /* Shrink children to necessary number, if any exist	*/
      /* If not, it's a leaf, so add itself to its subtree	*/

      if (n->children != 0) {
	(n->numnodes)++;	/* This node itself is no leaf	*/
	n->children = (nodeptr*)realloc( (void*) n->children,
					 n->numchildren * sizeof(nodeptr));
      } else {
	n->numleafs = 1;
	n->numnodes = 0;
      }
      n->procnum = procnum;	/* Remember procnum/cpu time if found	*/
      n->cputime = cputime;
      if (procnum > maxprocnum)	/* Update global maximum-proc.-index	*/
	  maxprocnum = procnum;

      return n;

    } else if ((c >= '0') && (c <= '9')) {

      /* Parse procedure number resp. CPU time */

      if (haveprocnum) {	/* Procnum read, now comes the CPU time */
	cputime = cputime * 10 + (long) (c - '0');
      } else {			/* Still in procnum			*/
	procnum = procnum * 10 + (int) (c - '0');
      }


    } else if (c == ':') {

      haveprocnum = 1;		/* End of procedure number seen		*/

    } else {

      printf("parse_tree: unexpected character '%c'\n", c);

    }

  }

  if (procnum > maxprocnum)	/* Update global maximum-proc.-index	*/
      maxprocnum = procnum;
  
  n->procnum = procnum;		/* Remember procnum/cpu time if found	*/
  n->cputime = cputime;
  return n;

}


/* Outputs the tree semi-graphically with the number of
 * leafs and nodes in the subtree. If verbose is true,
 * also output procedure number (as "A"..."Z") and CPU time used.
 * Setting procverbose to true results in subtree-node/leaf output
 * for each recursive procedure type.
 */
void print_tree(nodeptr n, int depth, char first,
		char verbose, char procverbose)
{
    int t;
    int i;
    int j;
    char any;
    char f;

    if (!first)
      for (t=0; t< depth; t++) {        /* Indent for depth     	*/
	  if ((verbose==1) && (procverbose==0)) {
	      printf("                   ");
	} else if ((verbose==0) && (procverbose==1)) {
	  printf("            ");
	  for (j=0; j<maxprocnum; j++)	/* Room for procverbose		*/
	      printf("            ");
	} else if ((verbose==1) && (procverbose==1)) {
	  for (j=0; j<maxprocnum; j++)	/* Room for procverbose		*/
	      printf("            ");
	} else {
	  printf("            ");
	}
      }
    
    printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
	   n->cputime, n->numnodes, n->numleafs);
    f=1;

    for (i=0; i<n->numchildren; i++) {
      print_tree(n->children[i], depth+1, f, verbose, procverbose);
      f=0; 	                   	/* Reset flag after 1st print	*/
    }
    if (!n->numchildren)
      printf("\n");             	/* New line if no children  	*/
    
}


/* Print tree graphically using ASCII output.
 * If verbose flag is set, print procedure number (A,B,...)
 * instead of node symbol.
 */
void print_tree_graphics(nodeptr n, int depth,
			 char first, char last, char* branches,
			 char verbose)
{
    int t;
    int i;
    char f;	/* Flag: True if first child of this node	*/
    char l;	/* Flag: True if last child of this node	*/
    char any;

    if (!first) {
      for (t=0; t < depth; t++)		/* Indent for depth	*/
	  if (verbose) {
	      printf("%c       ", branches[t]);
	  } else {
	      printf("%c  ", branches[t]);
	  }
    }
    
    f=1;


    for (i=0; i<n->numchildren; i++) {	/* Lengthen branch if any subbranches*/
	l = (char)(i >= n->numchildren-1);
	if (f) {
	    if (verbose) {		/* Procedure number as character*/
		if (n->threadid == -1)
		    printf("%c%c====--", (char)(n->procnum + 'A'-1),
			   (n->nodedone)?'d':'.');	/* Done or not	*/
		else
		    printf("%c%c%03d%c--", (char)(n->procnum + 'A'-1),
			   (n->nodedone)?'d':'.',	/* Done or not	*/
			   n->threadid,		/* Thread ID	*/
			   statechar(threads[n->threadid].state));
	    }
	    else
	        if ((last) && (!first))
		    printf("`--");	/* Last branch looks different	*/
	        else	
		    printf("+--");
	    branches[depth+1]='|';
	}				/* Set flag for last child	*/
	if (l)				/* Only reset after last child	*/
	    branches[depth+1]=' ';
	print_tree_graphics(n->children[i], depth+1,
			    f, l, branches, verbose);
	f=0;
    }
    if (!n->numchildren)		/* Print leaf correctly		*/
        if (verbose) {
	    if (n->threadid == -1)
		printf("%c%c====\n", (char)(n->procnum + 'A'-1),
		       (n->nodedone)?'d':'.');	/* Done or not	*/
	    else
		printf("%c%c%03d%c\n", (char)(n->procnum + 'A'-1),
		       (n->nodedone)?'d':'.',	/* Done or not	*/
		       n->threadid,		/* Thread ID	*/
		       statechar(threads[n->threadid].state));
	} else {
	    if ((last) && (!first))
	        printf("`-O\n"); 
	    else
	        printf("+-O\n"); 
	}

}


/* Compute depth of tree recursively
 * (a one-leaf-only tree has depth 1 but its node(leaf) will be at
 * level 0, i.e. the depths of the tree range from 0..tree_depth(root) !)
 */
int tree_depth(nodeptr n)
{
    int i;
    int d;
    int cd;

    d=1;				/* Own depth-increase		*/
    for (i=0; i<n->numchildren; i++) {
	cd = tree_depth(n->children[i]);
	if (cd+1 > d)			/* Remember maximum child depth	*/
	    d=cd+1;
    }
    return d;
   
}


/* Compute maximum branching degree of tree recursively
 */
int tree_degree(nodeptr n)
{
    int i;
    int d;
    int cd;

    d=n->numchildren;			/* Own degree			*/
    for (i=0; i<n->numchildren; i++) {
	cd = tree_degree(n->children[i]);
	if (cd > d)			/* Remember maximum child degree*/
	    d=cd;
    }
    return d;
   
}


/* Prints statistics on number of nodes/leafs in subtrees
 * at given depth immediately
 */
void print_subtrees_at_depth(nodeptr n, int depth,
			     char verbose, char procverbose)
{
    if (depth<=0) {
	printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
	       n->cputime, n->numnodes, n->numleafs);
    } else {
	int i;
	for (i=0; i<n->numchildren; i++)
	    print_subtrees_at_depth(n->children[i], depth-1,
				    verbose, procverbose);
    }
}


/* Collects statistics on number of nodes/leafs in subtrees
 * at given depth and returns pseudo-node with array of pointers to nodes.
 * info[1] gives number of nodes at exactly this depth,
 * info[2] ditto for leafs (info[0] is used internally)
 */
nodeptr subtrees_at_depth(nodeptr n, int depth)
{
    nodeptr statnode;
    int     t;
    
    statnode = (nodeptr) malloc(sizeof(node));
    statnode->numchildren = 0;		/* Initialize values		*/
    statnode->children = (nodeptr*)0;
    statnode->numnodes = 0;
    statnode->numleafs = 0;
    statnode->info     = (int*)malloc(3*sizeof(int));
    statnode->info[0]  = 0;	/* # of currently allocated children	*/
    statnode->info[1]  = 0;	/* # of nodes at exactly this depth	*/
    statnode->info[2]  = 0;	/* # of leafs ditto			*/

    get_subtrees_at_depth(n, depth, statnode);

    /* Shrink array of children to necessary size		*/
    
    if (statnode->children != 0) {
	statnode->children = (nodeptr*)realloc( (void*) statnode->children,
						statnode->numchildren *
						sizeof(nodeptr));
    }

    return statnode;
}


void get_subtrees_at_depth(nodeptr n, int depth, nodeptr statnode)
{
    int t;
    
    if (depth < 0)
	return;
    if (depth == 0) {		/* Depth found, add stats to node*/

      (statnode->numchildren)++;
      if (statnode->children == 0) {
	
	  /* Didn't have children yet, allocate some space	*/

	  statnode->children = (nodeptr*)malloc(CHILDNUM * sizeof(nodeptr));
	  statnode->info[0] = CHILDNUM;/* Remember current max.alloc.*/

      } else if (statnode->numchildren >= statnode->info[0]) {
	
	/* More children than already allocated, make room	*/

	statnode->info[0] = statnode->info[0] + CHILDNUM;
	statnode->children = (nodeptr*)realloc( (void*) statnode->children,
						statnode->info[0] *
						sizeof(nodeptr));
      }

      statnode->children[statnode->numchildren-1] = n;
      statnode->numleafs = statnode->numleafs +		/* Count leafs	*/
	                   n->numleafs;
      statnode->numnodes = statnode->numnodes +		/* Count nodes	*/
  	                   n->numnodes;
      if (n->numchildren == 0)
	  (statnode->info[2])++;      /* Count leafs at THIS depth only */
      else
	  (statnode->info[1])++;      /* Ditto nodes			*/

    } else {
	int i;
	for (i=0; i<n->numchildren; i++)
	    get_subtrees_at_depth(n->children[i], depth-1, statnode);
    }
}


/* Also initialize parleaf-status (all subtrees have non-parallel children -
 * Parleaf=1 means that the node isn't a "real" leaf but its subtrees
 * only contain nodes whose procedures are non-parallel)
 * Returns 1 if all subnodes includind the node itself are non-parallel,
 * 0 otherwise
 * Also collects statistics for procedures (number of nodes/leafs etc)
 */
char init_sub_node_status(nodeptr n, int depth)
{
    int  i;
    char pl;
    
    n->threadid = -1;
    n->nodedone = 0;
    n->depth    = depth;

    if (n->numchildren) {
	procs[n->procnum].nodes++;	/* Keep track of node/leaf stats*/
    } else {
	procs[n->procnum].leafs++;
    }
    
    pl = 1;
    for (i=0; i<n->numchildren; i++) {
	char c_pl;
	c_pl = init_sub_node_status(n->children[i], depth+1);
	if (c_pl == 0)
	    pl=0;
    }
    if ((pl) && (procs[n->procnum].par)) {/* Node itself is parallel 	*/
	pl = 0;				/*  but children are non-par.	*/
	n->parleaf = 1;
	procs[n->procnum].parleafs++;
    } else {
	n->parleaf = 0;
	if (procs[n->procnum].par) {
	    procs[n->procnum].parnodes++;
	}
    }
    return(pl);
}


/* Initialize (sub)tree with "no threads worked here yet" info
 * as well as parleaf status and procedure statistics
 */
void init_node_status(nodeptr n, int depth)
{
    char c;
    c = init_sub_node_status(n,depth);
}

/* Puts thread at end of thread queue (only for active or waiting
 * threads that are not bound to CPUs)
 * The calling procedure must make sure the thread meets those criteria.
 * Since threads that are active on some CPU (on_cpu=1) are not in the
 * queue anyway, they don't need to be removed when they change their
 * state to an inactive one.
 */
void enqueue_thread(int i) {

    if ((threads[i].on_cpu == 0) &&
	(threads[i].state != T_NONE) &&
	(threads[i].state != T_JOINED) &&
	(threads[i].state != T_DONE) ) {
	threadqlen++;
	threadq[threadqlen-1] = i;		/* Put thread in ready Q    */
    } else {
	printf("WARNING, CPU-bound/none/joined/done thread enqueued, ignoring it!\n");
    }
    
}


/* Take first thread from ready queue, makes it active on the given CPU,
 * and return its ID (-1 if no thread in Q).
 * (The thread queue only contains threads that are ACTIVE or WAITING)
 */
int dequeue_thread(int cpu) {
    int i;
    int t;
    
    if (threadqlen == 0) {		/* No threads at all, oops	*/
	return -1;
    }
    
    t = threadq[0]; 			/* Get 1st thread from queue	*/
    threads[t].on_cpu = 1;		/* Mark it active on given CPU	*/
    cputhreads[cpu] = t;
    
    for (i=0; i<threadqlen; i++) {	/* Delete thread from queue	*/
	threadq[i] = threadq[i+1];
    }
    threadqlen--;
    threadq[threadqlen] = -1;

    return(t);
}


/* "Starts" a new thread at the given node
 */
int new_thread(nodeptr n, char verbose)
{
    numthreads++;
    if (numthreads >= MAXTHREADS) {
	printf("ERROR, more than MAXTHREADS (%d) simulated threads\n",
	       MAXTHREADS);
	exit(333);
    }
    threads[numthreads].state = T_ACTIVE;	/* Thread working on node   */
    threads[numthreads].startnode = n;
    threads[numthreads].currentnode = n;
    threads[numthreads].nodesdone = 0;
    threads[numthreads].leafsdone = 0;
    threads[numthreads].subnodesdone = 0;
    threads[numthreads].subleafsdone = 0;
    threads[numthreads].on_cpu = 0;		/* Not scheduled on CPU yet */
    threads[numthreads].startdepth = n->depth;
    threads[numthreads].maxdepth = n->depth;
    n->threadid = numthreads;			/* Node "belongs" to thread */
    enqueue_thread(numthreads);			/* Put thread in ready queue*/

    numactive++;				/* New active threads count */

    if (verbose)
	printf("New thread, ID %d, Node %d, not on CPU yet, %d active now\n",
	       numthreads, n, numactive);
    
    return(numthreads);
}


/* Simulate one step (thread action) for one CPU
 */
void simulate_next(int cpu, nodeptr root, char verbose)
{
    nodeptr  n;
    int      id;
    int      i;
    thread  *t;
    int	     curthread;
    char     skip_nopar_procs;
    int      workdone;
    

    /* Examine next thread */

    curthread = cputhreads[cpu];	/* What is this CPU doing? 	*/

    if (verbose) {
	printf("\nnumthreads %d, numactive %d\n",
	       numthreads, numactive);
	for (i=0; i<cpunum; i++) {
	    if (cputhreads[i] >= 0)
	        printf("  CPU %2d: Thread %3d\n", i, cputhreads[i]);
	    else
	        printf("  CPU %2d: Thread ===\n", i);
	}
	for (i=0; i<=numthreads; i++) {
	    printf("  Thread %3d: state %c, start %3d, max %3d, %3d nodes, %3d leafs, %3d subnodes, %3d subleafs",
		   i, statechar(threads[i].state),
		   threads[i].startdepth, threads[i].maxdepth,
		   threads[i].nodesdone, threads[i].leafsdone,
		   threads[i].subnodesdone, threads[i].subleafsdone);
	    if (threads[i].on_cpu)
		printf(", on cpu\n");
	    else
		printf("\n");
	}
	printf("  Queue: ");
	for (i=0; i<threadqlen; i++) {
	    printf("%3d ", threadq[i]);
	}
	printf("\n");
    }

    if (curthread == -1) {		/* CPU idle, grab active thread	*/

	/* Select thread that's not on another CPU and active or waiting.
	 * The thread queue does that job for us, so get thread from there!
	 */
	curthread = dequeue_thread(cpu);
	if ((curthread != -1) && (verbose))
	    printf("Scheduled thread %d (state %c) on CPU %d\n",
		   curthread, statechar(threads[curthread].state),cpu);
    }
    
    if (curthread == -1) {
	if (verbose)
	    printf("No executable thread for CPU %d\n", cpu);
	return;
    }

    workdone = 0;		/* Keep track of work (not just waiting)*/
    
    t = &(threads[curthread]);

    
    /* Check thread's action */
    
    /* Loop while nodes belonging to procedures marked as "not parallel"
     * are skipped. Default: execute switch just once
     */
    skip_nopar_procs = 0;
    do {
	
    n = t->currentnode;

    if ((procs[n->procnum].par) ||	/* Update skip-status		    */
	(n->nodedone))	{		/* (stop skipping when node done)   */
	if ((skip_nopar_procs) && (verbose)) {
	    printf("   (no longer in skipping mode");
	    if (n->nodedone)
	      printf(" - node done!");
	    printf(")\n");
	}
        skip_nopar_procs = 0;
    } else {
        skip_nopar_procs = seqonestep;	/* Only skip if "one big step" set */
	if ((verbose) && (skip_nopar_procs)) {
	    printf("   (in non-parallel procedure skipping mode)\n");
	}
    }
    
    if ((!skip_nopar_procs) && (
	n->depth > t->maxdepth)) {	/* Keep track of max. depth reached */
	t->maxdepth = n->depth;		/* (only if not skipping!)	    */
    }

    if (verbose) {			/* Output tree and state if verbose */
	char branchtext[BUFSIZ];
	int  j;
	for (j=0; j<BUFSIZ; j++) branchtext[j] = ' ';
 	printf("CPU %d thread %d - state %c, depth %d, current node %d, start node %d\n",
	       cpu, curthread, statechar(t->state), n->depth, n, t->startnode);
	print_tree_graphics(root, 1, 0, 1, branchtext, 1);
    }

			  
    switch (t->state) {

    /* Waiting: Check if all child nodes are done (join child's thread
     *	    if it's done, skipping children after un-done ones).
     *      If all children are done, this node is done and the
     *	    thread returns to its parent node and becomes ACTIVE
     *	    again, unless we are already at its starting node,
     *	    in which case the thread is now DONE.
     */
    case T_WAITING: {
	char done = 1;
	i = 0;
	while ((i<n->numchildren) && (done)) {
	    if (!(n->children[i]->nodedone)) {	/* Subtree not done 	*/
		done = 0;		/* Skip remaining children	*/
	    } else {			/* Subtree's thrd done? Join it	*/
		if (threads[n->children[i]->threadid].state == T_DONE) {
		    threads[n->children[i]->threadid].state = T_JOINED;
		    if (strategy == S_KEEP)	/* keep_n -> one less	*/
			numactive--;
		    if (verbose)
			printf("W: Thread %d: Joined child thread %d, now active = %d\n",
			       curthread, n->children[i]->threadid, numactive);
		    workdone++;		/* Joining is work, too		*/
		}
	    }
	    i++;
	}
	if (done) {			/* All children done, continue	*/
	    workdone++;			/* Node done is work,too	*/
	    
	    n->nodedone = 1;		/* Current node is done now	*/
	    if (n->numchildren) {
		(t->nodesdone)++;	/* Update thread's performance	*/
		(t->subnodesdone)++;
	    } else {			/* (waiting implies threads, so	*/
		(t->leafsdone)++;	/*  we don't have to check for	*/
		(t->subleafsdone)++;	/*  parallel/serial procedures)	*/
	    }
	    if (n != t->startnode) {	/* Not at thread's root yet?	*/
		t->currentnode = n->parent; /* Continue at parent	*/
		t->state = T_ACTIVE;
		if (verbose)
		    printf("W: Thread %d: Child nodes done, parent work left, now active again\n",
			   curthread);
	    } else {			/* At root, thread done		*/
		t->state = T_DONE;
		if (strategy != S_KEEP)	/* depth/active_n -> one less	*/
		    numactive--;
		t->on_cpu = 0;		/* De-schedule thread from CPU	*/
		cputhreads[cpu] = -1;
		if (verbose)
		    printf("W: Thread %d: Child nodes done, at start node, done, yield CPU. Now active = %d\n",
			   curthread, numactive);
		skip_nopar_procs = 0;	/* Node done, skipping is over!	*/
	    }
	} else {			/* Still waiting...		*/
	    t->on_cpu = 0;		/* De-schedule thread from CPU	*/
	    cputhreads[cpu] = -1;
	    enqueue_thread(curthread);	/* Put into Q for later check	*/
	    if (verbose)
		printf("W: still waiting for children, yield CPU\n");
	    
	    /* This is no real work, so don't add to workdone
	     * Otherwise, idle threads get sheduled, see that nothing
	     * is to be done, and go to sleep again - that shouldn't
	     * be counted as work (ideally, sleeping threads should
	     * probably be woken up by some condition and not reside
	     * in the thread queue ...)
	     */
	}
	break;
    }
    /* Active: Examine next yet-untouched child and create a new
     *	   thread for it if the thread condition holds resp.
     *	   work on it directly.
     *	   If all children are touched, WAIT for them to finish.
     *	   If no children exist, return to parent resp. become DONE
     */
    case T_ACTIVE: {
	char isleaf;
	char childrendone;

	workdone++;			/* This thread did something	*/
	
	isleaf = (n->numchildren == 0);	/* Is this node a leaf?		*/
	childrendone = 1;
	if (!isleaf) {			/* Are all children done?	*/
	    for (i=0; i<n->numchildren; i++) {
		if (n->children[i]->nodedone == 0)
		    childrendone = 0;
	    }
	}
	if ((!isleaf) && (!childrendone)) {/* Has children left to work	*/

	    i = 0;
	    while ((i<n->numchildren) && (n->children[i]->threadid != -1)) {
		i++;			/* Look for children w/o thread	*/
	    }
	    if (i>= n->numchildren) {	/* All children busy, wait	*/
		t->state = T_WAITING;	/* (don't remove from Q yet)	*/
		if (verbose)
		    printf("A: Thread %d: All children busy, wait for them\n",
			   curthread);

	    } else {
		char strategy_ok;	/* Strategy says 'new thread'?	*/
		char parallel_ok;	/* Procedure to be parallelized?*/
		char nothread_ok;	/* No Nothread annotation?	*/

		if (strategy == S_DEPTH) {
		    strategy_ok = (n->depth < stratparam-1);
		} else if ((strategy == S_KEEP) || (strategy == S_ACTIVE)) {
		    strategy_ok = (numactive < stratparam * cpunum);
		} else if (strategy == S_FIRST) {
		    strategy_ok = (numthreads < stratparam * cpunum);
		} else if (strategy == S_ALWAYS) {
		    strategy_ok = 1;
		} else if (strategy == S_NEVER) {
		    strategy_ok = 0;
		}
		parallel_ok = (procs[n->children[i]->procnum].par);
		nothread_ok = (!procs[n->procnum].nothread[i]);
		
		/* Activate next child if the thread strategy allows for it
		 * AND the child node's procedure is to be parallelized at all
		 * AND the current branch is not annotated as "nothread" for
		 * the current procedure
		 */
		if (strategy_ok && parallel_ok && nothread_ok) {

		    new_thread(n->children[i], verbose);/* Spawn thread	*/
		    if (verbose)
 			printf("A: Thread %d: Spawned thread for child %d\n",
 			       curthread, i);

		} else {

		    t->currentnode = n->children[i];	/* Doityourself */
		    t->currentnode->threadid = curthread;
		    
		    /* Child is not parallel? Start skipping-sequential-
		     * mode if "-e" enabled ("one big step for seq.procs")
		     */
		    if (!parallel_ok) {
			skip_nopar_procs = seqonestep;
		    }
		    
		    if (verbose) {
			if (parallel_ok) {
			    printf("A: Thread %d: Working on child %d ourselves",
				   curthread, i);
			    if (strategy_ok && (! nothread_ok)) {
				printf(" (nothread annotation)");
			    }
			    printf("\n");
			} else {
			    printf("A: Thread %d: Skipping over non-parallel procedure of child %d\n",
				curthread, i);
			}
		    }
		}
	    }
	} else {			/* No children, return to parent*/
	    n->nodedone = 1;		/* Current node is done now	*/
	    if (n->numchildren) {
		(t->subnodesdone)++;	/* Update # of nodes/leafs	*/
	    } else {			/*  regardess of parallel status*/
		(t->subleafsdone)++;
	    }
	    if (!skip_nopar_procs) {	/* Now for the parallel-only:	*/
		if ((n->numchildren) && (!n->parleaf)) {
		    (t->nodesdone)++;	/* Update thread's performance	*/
		} else {		/* (only of not skipping)	*/
		    (t->leafsdone)++;	/* Recognize parallel "leafs"	*/
		}
	    }
	    if (n != t->startnode) {	/* Not at thread's root yet?	*/
		n = n->parent;		/* Continue at parent		*/
		t->currentnode = n;
		if (verbose)
		    printf("A: Thread %d: Current node done, working on parent\n",
			   curthread);
	    } else {			/* At root, thread done		*/
		t->state = T_DONE;
		if (strategy != S_KEEP)	/* depth/active_n -> one less	*/
		    numactive--;
		t->on_cpu = 0;		/* De-schedule thread from CPU	*/
		cputhreads[cpu] = -1;
		if (verbose)
		    printf("A: Thread %d: Current node done, was start node, done. Yield CPU\n",
			   curthread);
	    }
	}
	break;
    }
    /* Done: Don't do anything
     */
    case T_DONE: {
	t->on_cpu = 0;			/* De-schedule thread from CPU	*/
	cputhreads[cpu] = -1;
	if (verbose)
	    printf("Warning - de-scheduling DONE thread!\n");
	break;
    }
    /* Joined: Don't do anything
     */
    case T_JOINED: {
	t->on_cpu = 0;			/* De-schedule thread from CPU	*/
	cputhreads[cpu] = -1;
	if (verbose)
	    printf("Warning - de-scheduling JOINED thread!\n");
	break;
    }
    /* None: shouldn't happen, threads start ACTIVE
     *       and never return to NONE
     */
    case T_NONE: {
	printf("ERROR - NONE state for thread %d\n", curthread);
	exit(444);
	break;
    }
    default: {
	printf("ERROR - unknown state for thread %d\n", curthread);
	exit(555);
	break;
    }
	    
    } /* switch */
    
    } while (skip_nopar_procs);

    if (workdone)		/* Anything done (not just waiting)?	*/
        cpusbusy[simsteps]++;	/* Keep track of CPU busy status	*/

}

/* Perform the simulation
 */
void simulate(nodeptr root, char verbose)
{
    nodeptr n;
    int     i;
    int     myid;

    init_node_status(root,0);		/* Make sure tree is "un-worked"*/
    for (i=0; i<MAXTHREADS; i++) {	/* Initialize thread array	*/
	threads[i].state       = T_NONE;
	threads[i].on_cpu      = 0;
	threads[i].startnode   = NULLPTR;
	threads[i].currentnode = NULLPTR;
	threadq[i] = 0;
    }
    for (i=0; i<cpunum; i++) {		/* Init current cpu threads	*/
	cputhreads[i] = -1;
    }
    threadqlen    = 0;
    numthreads    = -1;			/* Initial thread counter	*/
    numactive     = 0;			/* (increased to 0 by new_thr!)	*/
    myid          = new_thread(root,verbose);/* 1st thread starts @root	*/
    myid	  = dequeue_thread(0);	/* Start thread on CPU 0	*/

    /* 1st thread doesn't count for active#, doesn't in real parallel
     * program, either!
     */
    numactive=0;

    /* Keep simulating on all CPUs until root node is done */
    
    while (! root->nodedone) {

	/* Also simulate Solaris abort if more than 3000 threads in Q
	 */
	if (threadqlen >= 3000) {
	    printf("WARNING, more than 3000 threads at once, aborting!\n");
	    return;
	}

	cpusbusy[simsteps] = 0;		/* Is incremented in simulation	*/
	
	for (i=0; i<cpunum; i++) {
	    simulate_next(i, root, verbose);
/* 	    procswitch++; */
/* 	    if ((procswitch%17) == 0) {	/* Simulate process switch	*/
/* 		if (cputhreads[i] != -1) { */
/* 		    threads[cputhreads[i]].on_cpu = 0; */
/* 		    enqueue_thread(cputhreads[i]); */
/* 		    cputhreads[i] = -1; */
/* 		} */
/* 		if (verbose) printf("switch!\n"); */
/* 	    } */
	}

	if (verbose) {
	    printf("%d CPUs busy during the last simulation step\n",
		   cpusbusy[simsteps]);
	}
	simsteps++;
	if (simsteps >= simstepsalloc) {
	    simstepsalloc = simstepsalloc + 1000;
	    cpusbusy = (int*) realloc((void*) cpusbusy,
				     simstepsalloc * sizeof(int));
	}
    }
    
}


/* Compute average of integer array
 */
double avg(int *array, int n)
{
    int    i;
    double a;

    a=0.0;

    for (i=0; i<n; i++)
	a = a + (double)array[i];

    return a/(double)n;
    
}


/* Given the array and the average, computes the deviation
 */
double dev(int *array, int n, double a)
{
    int    i;
    double d;

    d=0.0;

    for (i=0; i<n; i++)
	d = d + ((double)array[i]-a)*((double)array[i]-a);

    return sqrt(d/(double)n);

}


/* Heuristics: Evaluate result
 */
void heuristics(nodeptr root) {
    double cpupercent = 100.0 / (double) cpunum; /* Percentage of 1 CPU	*/
    
    int    *numnodes;		/* Number of parallel nodes computed	*/
    int    *numleafs;		/* Ditto leafs				*/
    int    *numsubnodes;	/* Including sequential leafs/nodes	*/
    int    *numsubleafs;
    double avgnumnodes;		/* Averages and deviations		*/
    double devnumnodes;
    double avgnumleafs;
    double devnumleafs;
    double avgnumsubnodes;
    double devnumsubnodes;
    double avgnumsubleafs;
    double devnumsubleafs;
    int    minnumnodes;		/* Minumum, median and maximum		*/
    int    mednumnodes;
    int    maxnumnodes;
    int    minnumleafs;
    int    mednumleafs;
    int    maxnumleafs;
    int    minnumsubnodes;
    int    mednumsubnodes;
    int    maxnumsubnodes;
    int    minnumsubleafs;
    int    mednumsubleafs;
    int    maxnumsubleafs;
    double numnodespercent;	/* Percentage of whole tree	*/
    double numleafspercent;
    double numsubnodespercent;
    double numsubleafspercent;
    int i;

    numnodes    = (int*) malloc ((numthreads+1) * sizeof(int));
    numleafs    = (int*) malloc ((numthreads+1) * sizeof(int));
    numsubnodes = (int*) malloc ((numthreads+1) * sizeof(int));
    numsubleafs = (int*) malloc ((numthreads+1) * sizeof(int));

    /* Copy thread data to arrays, sort it, and compute statistics */
    
    for (i=0; i<=numthreads; i++) {
	numnodes[i]    = threads[i].nodesdone;
	numleafs[i]    = threads[i].leafsdone;
	numsubnodes[i] = threads[i].subnodesdone;
	numsubleafs[i] = threads[i].subleafsdone;
    }
    intsort(numnodes, numthreads+1);
    intsort(numleafs, numthreads+1);
    intsort(numsubnodes, numthreads+1);
    intsort(numsubleafs, numthreads+1);

    minnumnodes    = numnodes[0];
    mednumnodes    = numnodes[numthreads/2];
    maxnumnodes    = numnodes[numthreads];
    minnumleafs    = numleafs[0];
    mednumleafs    = numleafs[numthreads/2];
    maxnumleafs    = numleafs[numthreads];
    minnumsubnodes = numsubnodes[0];
    mednumsubnodes = numsubnodes[numthreads/2];
    maxnumsubnodes = numsubnodes[numthreads];
    minnumsubleafs = numsubleafs[0];
    mednumsubleafs = numsubleafs[numthreads/2];
    maxnumsubleafs = numsubleafs[numthreads];

    avgnumnodes    = avg(numnodes, numthreads+1);
    devnumnodes    = dev(numnodes, numthreads+1, avgnumnodes);
    avgnumleafs    = avg(numleafs, numthreads+1);
    devnumleafs    = dev(numleafs, numthreads+1, avgnumleafs);
    avgnumsubnodes = avg(numsubnodes, numthreads+1);
    devnumsubnodes = dev(numsubnodes, numthreads+1, avgnumsubnodes);
    avgnumsubleafs = avg(numsubleafs, numthreads+1);
    devnumsubleafs = dev(numsubleafs, numthreads+1, avgnumsubleafs);

    printf("\nThread statistics for %d threads:\n", numthreads+1);
    
    printf(" nodes:    %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
	   minnumnodes, mednumnodes, maxnumnodes,
	   avgnumnodes, devnumnodes);
    if (avgnumnodes > 0)
	printf(" (%6.3f)\n", devnumnodes/avgnumnodes);
    else
	printf("\n");
    
    printf(" leafs:    %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
	   minnumleafs, mednumleafs, maxnumleafs,
	   avgnumleafs, devnumleafs);
    if (avgnumleafs > 0)
	printf(" (%6.3f)\n", devnumleafs/avgnumleafs);
    else
	printf("\n");
    
    printf(" subnodes: %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
	   minnumsubnodes, mednumsubnodes, maxnumsubnodes,
	   avgnumsubnodes, devnumsubnodes);
    if (avgnumsubnodes > 0)
	printf(" (%6.3f)\n", devnumsubnodes/avgnumsubnodes);
    else
	printf("\n");

    printf(" subleafs: %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
	   minnumsubleafs, mednumsubleafs, maxnumsubleafs,
	   avgnumsubleafs, devnumsubleafs);
    if (avgnumsubleafs > 0)
	printf(" (%6.3f)\n", devnumsubleafs/avgnumsubleafs);
    else
	printf("\n");
    
    numnodespercent    = 100.0 * (double)maxnumnodes/(double)root->numnodes;
    numleafspercent    = 100.0 * (double)maxnumleafs/(double)root->numleafs;
    numsubnodespercent = 100.0 * (double)maxnumsubnodes/(double)root->numnodes;
    numsubleafspercent = 100.0 * (double)maxnumsubleafs/(double)root->numleafs;
    
    printf("Largest values: %5d nodes (%6.2f%%), %5d leafs (%6.2f%%),",
	   maxnumnodes, numnodespercent,
	   maxnumleafs, numleafspercent);
    printf(" %5d subnodes (%6.2f%%), %5d subleafs (%6.2f%%)\n",
	   maxnumsubnodes, numsubnodespercent,
	   maxnumsubleafs, numsubleafspercent);
	   

    printf("\nHeuristics:\n");
    
    printf("Not less threads than CPUs                        : ");
    if (numthreads+1 >= cpunum) printf("OK\n");
    else printf("not\n");
    
    printf("Node percentage of largest <= 1 CPU percentage    : ");
    if (numnodespercent <= cpupercent) printf("OK\n");
    else printf("not\n");
    
    printf("Leaf percentage of largest <= 1 CPU percentage    : ");
    if (numleafspercent <= cpupercent) printf("OK\n");
    else printf("not\n");
    
    printf("Subnode percentage of largest <= 1 CPU percentage : ");
    if (numsubnodespercent <= cpupercent) printf("OK\n");
    else printf("not\n");
    
    printf("Subleaf percentage of largest <= 1 CPU percentage : ");
    if (numsubleafspercent <= cpupercent) printf("OK\n");
    else printf("not\n");

    printf("Each CPU used at least 75%% of the time            : ");
    if (allcpuload / (double) cpunum >= 75) printf("OK\n");
    else printf("not\n");
    
}


/* HACK: kill all children whose procnum (proc type) is
 * different from the parent's - 1st step towards real procedure-type
 * handling ;)
 */
void hack_kill_other_node_types(nodeptr n)
{
    int      i,j;
    int      remaining;
    nodeptr *newchildren;

    remaining = n->numchildren;
    for (i=0; i<n->numchildren; i++) {
	if (n->children[i]->procnum != n->procnum) {
	    n->children[i] = (nodeptr) NULL;
	    remaining--;
	}
    }
    if (remaining != n->numchildren) {
	j=0;
	if (remaining > 0) {
	    newchildren = (nodeptr*)malloc(remaining * sizeof(nodeptr));
	    for (i=0; i<n->numchildren; i++) {
		if (n->children[i]) {
		    newchildren[j] = n->children[i];
		    j++;
		}
	    }
	}
	n->numchildren = remaining;
	n->children = newchildren;
    }
    for (i=0; i<n->numchildren; i++) {
	hack_kill_other_node_types(n->children[i]);
    }
    
}


void print_strat() {
      if (strategy == S_KEEP) {
	  printf("KEEP %d", stratparam);
      } else if (strategy == S_ACTIVE) {
	  printf("ACTIVE %d", stratparam);
      } else if (strategy == S_ALWAYS) {
	  printf("ALWAYS");
      } else if (strategy == S_FIRST) {
	  printf("FIRST %d", stratparam);
      } else if (strategy == S_NEVER) {
	  printf("NEVER");
      } else if (strategy == S_DEPTH) {
	  printf("DEPTH %d", stratparam);
      } else {
	  printf("???");
      }
      printf(" on %d CPUs", cpunum);
}


main(int argc, char **argv)
{
  char    verbose        = 0;
  
  nodeptr root;
  nodeptr *subroot;
  char   *inputtext = (char*) malloc(1);
  char    inputbuf[TEXTCHUNKSIZE];
  int     inputlen = 0;
  char    branchtext[BUFSIZ];
  int     l;
  int     i,j;
  int     depth;
  int	  degree;
  nodeptr n;
  int     maxparprocs;
  long    busysum;


  /* Initializations */

  strategy     = S_KEEP;	/* Simulate KEEP_N Strategy		*/
  cpunum       = 4;		/* .. and 4 CPUs			*/
  stratparam   = 3;		/* .. and 3 Threads per CPU		*/
  for (j=0; j<MAXPROCS; j++) {
      int k;
      procs[j].par = 1;		/* All procedures parallel initially	*/
      for (k=0; k<MAXDEGREE; k++)
	procs[j].nothread[k] = 0;/* No nothread setting for any degree	*/
      procs[j].nodes    = 0;
      procs[j].leafs    = 0;	/* Init counters			*/
      procs[j].parnodes = 0;
      procs[j].parleafs = 0;
  }
  simsteps = 0;
  simstepsalloc = 1000;		/* Simulation steps (initial array)	*/
  cpusbusy  = (int*) malloc (simstepsalloc * sizeof(int));

  /* Parse command line */
      
  for (i=1; i<argc; i++) {
      if (!strcmp(argv[i],"-v")) {
	  printf("Verbose output\n");
	  verbose = 1;
      } else if (!strcmp(argv[i],"-q")) {
	  quietmode = 1;
      } else if (!strcmp(argv[i],"-e")) {
	  seqonestep = 1;
      } else if (!strcmp(argv[i],"-A")) {
	  strategy = S_ALWAYS;
      } else if (!strcmp(argv[i],"-a")) {
	  strategy = S_ACTIVE;
	  if ((i<argc-1) && (atoi(argv[i+1])>0)) {
	      stratparam = atoi(argv[i+1]);
	      i++;
	  }
      } else if (!strcmp(argv[i],"-d")) {
	  strategy = S_DEPTH;
	  if ((i<argc-1) && (atoi(argv[i+1])>0)) {
	      stratparam = atoi(argv[i+1]);
	      i++;
	  }
      } else if (!strcmp(argv[i],"-f")) {
	  strategy = S_FIRST;
	  if ((i<argc-1) && (atoi(argv[i+1])>0)) {
	      stratparam = atoi(argv[i+1]);
	      i++;
	  }
      } else if (!strcmp(argv[i],"-k")) {
	  strategy = S_KEEP;
	  if ((i<argc-1) && (atoi(argv[i+1])>0)) {
	      stratparam = atoi(argv[i+1]);
	      i++;
	  }
      } else if (!strcmp(argv[i],"-n")) {
	  strategy = S_NEVER;
      } else if (!strcmp(argv[i],"-s")) {
	  if (i==argc-1) {
	      printf("Error, -s needs argument (strategy A, a, d, f, k or n)\n");
	      exit(11);
	  }
	  i++;
	  if (!strcmp(argv[i],"A")) {
	      strategy = S_ALWAYS;
	  } else if (!strcmp(argv[i],"a")) {
	      strategy = S_ACTIVE;
	  } else if (!strcmp(argv[i],"d")) {
	      strategy = S_DEPTH;
	  } else if (!strcmp(argv[i],"f")) {
	      strategy = S_FIRST;
	  } else if (!strcmp(argv[i],"k")) {
	      strategy = S_KEEP;
	  } else if (!strcmp(argv[i],"n")) {
	      strategy = S_NEVER;
	  } else {
	      printf("Error, strategy argument must be 'A', 'a', 'd', 'f', 'k' or 'n')\n");
	      exit(12);
	  }
	  if ((i<argc-1) && (atoi(argv[i+1])>0)) {
	      stratparam = atoi(argv[i+1]);
	      i++;
	  }
      } else if (!strcmp(argv[i],"-c")) {
	  if (i==argc-1) {
	      printf("Error, -c needs argument (number of processors)\n");
	      exit(22);
	  }
	  cpunum = atoi(argv[i+1]);
	  i++;
      } else if (!strcmp(argv[i],"-p")) {
	  if (i==argc-1) {
	      printf("Error, -p needs argument (strategy parameter)\n");
	      exit(33);
	  }
	  stratparam = atoi(argv[i+1]);
	  i++;
      } else if (!strcmp(argv[i],"-t")) {
	  if (i==argc-1) {
	      printf("Error, -t needs argument (parallel procedures)\n");
	      exit(44);
	  }
	  for (j=0; j<MAXPROCS; j++) {	/* No procs parallel now	*/
	      procs[j].par = 0;
	  }
	  for (j=0; j<strlen(argv[i+1]); j++) {	/* Explicitly mark	*/
	      procs[argv[i+1][j]-'A'+1].par = 1;
	  }
	  i++;
      } else if (!strcmp(argv[i],"-b")) {
	  int pn;
	  int nt;
	  if (i==argc-1) {
	      printf("Error, -b needs argument (procedure and branch#)\n");
	      exit(55);
	  }
	  pn = argv[i+1][0] - 'A'+1;		/* Procedure number	*/
	  nt = atoi(argv[i+1] + 1);		/* Branching degree	*/
	  procs[pn].nothread[nt] = 1;
	  i++;
      } else {
	  printf("%s - simulate effects of thread generation strategies for\n",
		 argv[0]);
	  printf("     recursion tree read from stdin\n");
	  printf("Program options:\n");
	  printf("    -s A	Simulate always strategy (also '-A')\n");
	  printf("    -s a [n]	Simulate active_n strategy (also '-a [n]')\n");
	  printf("    -s d [n]	Simulate depth_n strategy (also '-d [n]')\n");
	  printf("    -s f [n]	Simulate first_n strategy (also '-f [n]')\n");
	  printf("    -s k [n]	Simulate keep_n strategy (also '-k [n]', default)\n");
	  printf("    -s n	Simulate never strategy (also '-n')\n");
	  printf("    -c C	Simulate C CPUs (default = 4)\n");
	  printf("    -p N	Strategy parameter: N*C threads with keep/active\n");
	  printf("    		resp. depth N (default = 3)\n");
	  printf("    -t ABCD..	Only execute procedure type A/B/... in parallel\n");
	  printf("      	(A=1st recursive procedure etc, default=all)\n");
          printf("    -b Ax	Don't create thread for branch #x of proc. A\n");
	  printf("		(A as above, branches count starting from 0)\n");
	  printf("    -e	Execute non-parallel parts in one single step\n");
	  printf("    -q	Quiet output, don't print final tree with threads\n");
	  printf("    -v	Verbose output\n");
	  printf("Output format for one node: PDTTTS\n");
	  printf("    P   = Node's procedure (A,B,...)\n");
	  printf("    D   = Node's state ('.' incomplete / 'd' done)\n");
	  printf("    TTT = Thread ID working on that node ('===' -> none)\n");
	  printf("    S   = Thread's status (None, Active, Waiting, Done, Joined)\n");
	  printf("\n");
	  exit(123);
      }
  }
  maxparprocs = 0;		/* Check maximum proc.number to be parall. */
  for (j=0; j<MAXPROCS; j++) {
      if (procs[j].par)
	  maxparprocs=j;
  }	
  if (verbose) {
      printf("Simulating ");
      print_strat();
      printf("\n");
      if (maxparprocs >= MAXPROCS-1) {
	  printf("All procedures simulated as parallelized\n");
      } else {
	  printf("Procedures simulated as parallelized: ");
	  for (j=0; j<=maxparprocs; j++) {
	      if (procs[j].par)
		  printf("%c ", (char)j+'A'-1);
	  }
	  printf("\n");
      }
      for (i=0; i<MAXPROCS; i++) {
	  for (j=0; j<MAXDEGREE; j++) {
  	      if (procs[i].nothread[j]) {
		  printf("'nothread' annotation for procedure %c, branching degree %d\n",
			 (char)i+'A'-1, j);
	      }
	  }
      }
      if (seqonestep) {
	  printf("Sequential procedures executed in single steps, too\n");
      }
  }

  if (stratparam <= 0) {
      printf("ERROR - strategy parameter must be > 0\n");
      exit(99);
  }

  /* Read tree from stdin */
 
  *inputtext = '\0';
  while (feof(stdin) == 0) {
      if (fgets(inputbuf, TEXTCHUNKSIZE, stdin)) {
	  l = strlen(inputbuf);
	  inputlen = inputlen + l;
	  inputtext = (char*)realloc(inputtext, inputlen);
/*	  printf("read %d bytes, chunk size = %d\n", l, TEXTCHUNKSIZE);*/
	  strcat(inputtext, inputbuf);
      }
  }
  if (strlen(inputtext)>1)
      text = inputtext+1;	/* Skip first dummy-root-node		*/
  else
      text = inputtext;

  maxprocnum = 0;		/* Init maximum procedure number	*/
  root = parse_tree();

  /* hack_kill_other_node_types(root);	/* HACK */
  
  if (verbose) {
      printf("\nTree indented = (nodes/leafs)\n");
      print_tree(root, 0, 1, 0, 0);
      for (i=0; i<BUFSIZ; i++) {
	  branchtext[i] = ' ';
      }
      printf("\nTree graphically (procedures as characters) =\n");
      print_tree_graphics(root, 1, 0, 1, branchtext, 1);
  }
  
  depth  = tree_depth(root);
  degree = tree_degree(root);
  if (verbose) {
      printf("\nTree depth  = %d\n", depth);
      printf("Tree degree = %d\n", degree);
      if (root->cputime > 0) {
	  printf("ms per node = %-12.6f\n",
		 (double)root->cputime / (double)root->numnodes);
	  printf("ms per leaf = %-12.6f\n",
		 (double)root->cputime / (double)root->numleafs);
      } else {
	  printf("no timings given\n");
      }
      printf("Maximum of %d different procedures in tree\n", maxprocnum);
  }
  if (degree>MAXDEGREE) {
    printf("ERROR - branching degree is %d, but only %d maximum defined\n",
	   degree, MAXDEGREE);
    exit(88);
  }

  simulate(root, verbose);	/* Perform the simulation		*/

  if (!quietmode) {
      printf("\nResult:\n\n");	/* Output resulting tree graphics	*/
      for (i=0; i<BUFSIZ; i++) {
	  branchtext[i] = ' ';
      }
      print_tree_graphics(root, 1, 0, 1, branchtext, 1);

      for (i=0; i<=numthreads; i++) {	/* Output all thread stats	*/

	  printf("  Thread %3d: start %3d, max %3d, %3d nodes, %3d leafs, %3d subnodes, %3d subleafs\n",
		 i,
		 threads[i].startdepth, threads[i].maxdepth,
		 threads[i].nodesdone, threads[i].leafsdone,
		 threads[i].subnodesdone, threads[i].subleafsdone);
      }
  }

  printf("\nSimulation for ");
  print_strat();
  printf(" done. %d threads generated over all\n", numthreads);

  if (threadqlen >= 3000) {
      printf("Simulation ABORTED after %d steps, more than 3000 threads at once!\n", simsteps);
      exit(32);
  }

  busysum = 0;
  for (j=0; j<simsteps; j++) {
      busysum = busysum + cpusbusy[j];
  }
  allcpuload = ((double) busysum) / ((double) simsteps) * 100.0;
  printf("%d simulation steps elapsed, CPU utilization = %3.0f%%\n",
	 simsteps, allcpuload);

  if (seqonestep) {
      printf("Sequential procedures executed in one single step\n");
  }
      
  if (maxparprocs >= MAXPROCS-1) {	/* Output parallelization info	*/
      printf("All procedures simulated as parallelized\n");
  } else {
      printf("Procedures simulated as parallelized: ");
      for (j=0; j<=maxparprocs; j++) {
	  if (procs[j].par)
	      printf("%c ", (char)j+'A'-1);
      }
      printf("\n");
  }
  for (i=0; i<MAXPROCS; i++) {
      for (j=0; j<MAXDEGREE; j++) {
	  if (procs[i].nothread[j]) {
	      printf("'nothread' annotation for procedure %c, branching degree %d\n",
		     (char)i+'A'-1, j);
	  }
      }
  }
  printf("\n");
  for (i=0; i<MAXPROCS; i++) {
      if (procs[i].nodes + procs[i].leafs +	/* Proc. used at all? */
	  procs[i].parnodes + procs[i].parleafs > 0) {
	  printf("Procedure %c: %5d nodes, %5d leafs, %5d parnodes, %5d parleafs\n",
		 (char)i+'A'-1, procs[i].nodes, procs[i].leafs,
		 procs[i].parnodes, procs[i].parleafs);
      }
  }

  heuristics(root);

  exit(0);
}

