/*
 * tree_parser.c
 *
 * Parses tree input from statistics generation and generates
 * in-memory tree. Performs statistical analysis and depth
 * thread generation heuristics.
 *
 *
 * Usage:   tree_parser [-v] < treefile
 *           -d  output detailed information
 *           -g  output graphical tree
 *           -gv ditto  verbose (procedure name tags)
 *           -s  output statistics
 *           -sv ditto  verbose (sizes of subtrees)
 *           -t  output tree
 *           -tv ditto  verbose (timings etc)
 *	    treefile	ASCII representation of recursion tree, e.g.
 *			"(()())" for simple binary tree, may be annotated
 *		        with procedure numbers and subtree timings 
 *
 *
 * Stefan Haenssgen  18-aug-97
 *
 *
 * 18-aug-97	1st version, parse parenthesis, count # of leafs
 * 20-aug-97	Also register # of nodes in subtree (excluding leafs)
 *		Take input from stdin
 *		Compute depth of tree recursively
 *		subtrees_at_depth for statistics on number and size
 *		 of subtrees
 *		Compute min, median and max of leafs & nodes in subtree
 * 22-aug-97	Also compute average and standard deviation
 *		Also compute number of nodes/leafs at exactly one depth
 *		 as part of subtree computation
 *		Compute maximum branching degree
 *		Consistently print nodes first, then leafs
 * 25-aug-97	Also compute subtree statistics for ALL nodes
 *		Extra function for average and deviation
 *		Show extended output (tree graphics etc) only when
 *		 "-v" switch set
 *		Keep all information on subtrees, indexed by depth
 *		Added heuristics for choice of DEPTH_N parameter
 * 28-aug-97	Debugged graphical tree output by adding "last" parameter
 * 03-sep-97	Recognize and parse additional procedure number and
 *		 CPU time (as ppp:ttt after opening parenthesis)
 *		Added this data to struct
 *		Added verbose switch to output procedures
 *		 to show procnum and CPU time
 * 04-sep-97	Verbose switch for subtrees-at-depth, too
 *		Output function as "A".."Z" if procnum defined (>0), "@" else
 * 05-sep-97	Added program options:
 *		 -d  debugging output
 *		 -t  output tree
 *		 -tv ditto verbose (timings etc)
 *		 -g  output graphical tree
 *		 -gv ditto verbose (procedure name tags)
 *		 -s  output statistics
 *		 -sv ditto verbose (sizes of subtrees)
 * 08-sep-97	Estimate granularity as total time/number of leafs
 *		 (add ":milliseconds_runtime" after first "(" in tree
 *		 manually for hand timing)
 *		Compute leafs/nodes for each procedure type separately
 *		 (procleafs/procnodes) in extra pass through tree - "leaf"
 *		 as no other procedures of the same type in its subtree
 *		Added procverbose switch to print_tree and subtree output
 *		 (procnum 0 is never used, so just skip it. maxprocnum
 *		 0 means no procnums given)
 *		Also generate proc-specific numbers in subtrees-at-depth
 * 10-sep-97	Debugged first-n-children output from "-sv"
 *		Added procnum_subtrees_till_depth to collect all
 *		 subtrees of specified procedure type until given depth
 * 03-oct-97	Updated comments/description at start of file
 *
 *		 !!! time etc also for each procedure type seperately
 *		
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define TEXTCHUNKSIZE 16384 /* Size of text string chuncs (4 allocation)*/

#define CHILDNUM 10	/* Default starting number of children for nodes*/

typedef struct _node *nodeptr;

typedef struct _node {
  int      numchildren;	/* Number of children this node has		*/
  nodeptr *children;	/* Array of children for this node		*/
  int 	   numnodes;	/* Number of nodes (without leafs) in subtree	*/
  int      numleafs;	/* Number of leafs in this subtree		*/
  int      procnum;	/* Number code for the node's procedure		*/
  long     cputime;	/* Used CPU time for this subtree in millisec.	*/
  int	  *procnodes;	/* Number of leafs for each procedure type	*/
  int	  *procleafs;	/* Ditto for number of nodes (from 0..maxpn-1)	*/
  int     *info;	/* Any additional information (size, ...)	*/
} node;

#define NULLPTR (nodeptr)NULL;

char *text;		/* Global pointer to current input position	*/

int maxprocnum;		/* Global maximum number of proc-indices found	*/
                        /* Zero here means no proc-numbers were found   */
                        /* at all. procnums range 1..mp, arrays 0..mp-1	*/


/* Forward declarations */

nodeptr parse_tree();
void    print_tree(nodeptr n, int depth, char first,
		   char verbose, char procverbose);
int     tree_depth(nodeptr n);
void    print_subtrees_at_depth(nodeptr n, int depth,
				char verbose, char procverbose);
nodeptr subtrees_at_depth(nodeptr n, int depth);
void    get_subtrees_at_depth(nodeptr n, int depth, nodeptr statnode);
nodeptr procnum_subtrees_till_depth(nodeptr n, int procnum, int depth);
void    get_procnum_subtrees_till_depth(nodeptr n, int procnum, int depth,
					nodeptr statnode);
void    print_tree_graphics(nodeptr n, int depth,
	   		    char first, char last, char* branches,
			    char verbose);


/* For quicksort
 */
static int intcompare(int *i, int *j)
{
    if (*i > *j)
	return (1);
    if (*i < *j)
	return (-1);
    return (0);
}


/* Sorts array of integers using library quicksort
 */
void intsort(int *a, int len)
{
  qsort((char *) a, len, sizeof(int), intcompare);
}


/* Parse global input text recursively:
 * "(" opens a new node and descends into it, ")" returns from a node
 */
nodeptr parse_tree() {
  char     c;			/* Current character in input	*/
  int      count;		/* Count children		*/
  nodeptr  n;			/* This node			*/
  nodeptr *cs;			/* Child pointer for this node	*/
  int      maxchildren;		/* Currently allocated #children*/
  int      procnum = 0;		/* Procedure code-number 	*/
  char	   haveprocnum = 0;	/* True procnum	already read	*/
  long     cputime = 0;		/* ms CPU Time spent in subtree	*/


  n = (nodeptr) malloc(sizeof(node));
  n->numchildren = 0;		/* Init values			*/
  n->children = (nodeptr*)0;
  n->numleafs = 0;
  n->numnodes = 0; 
  n->procnum  = 0;
  n->cputime  = 0;
  n->info = NULL;
  n->procleafs = (int*) NULL;	/* Procedure-specific counters	*/
  n->procnodes = (int*) NULL;	/*  are filled in later		*/
  

  while (c = *(text++)) {	/* Read next character		*/

    if (c == '(') {		/* New child, descend		*/

      (n->numchildren)++;
      if (n->children == 0) {
	
	/* Didn't have children yet, allocate some space	*/

	n->children = (nodeptr*)malloc(CHILDNUM * sizeof(nodeptr));
	maxchildren = CHILDNUM;	/* Remember current max.alloc.	*/

      } else if (n->numchildren >= maxchildren) {
	
	/* More children than already allocated, make room	*/

	maxchildren = maxchildren + CHILDNUM;
	n->children = (nodeptr*)realloc( (void*) n->children,
					 maxchildren * sizeof(nodeptr));
      }

      n->children[n->numchildren-1] = parse_tree();
      n->numleafs = n->numleafs +		/* Count leafs	*/
	            n->children[n->numchildren-1]->numleafs;
      n->numnodes = n->numnodes +		/* Count nodes	*/
	            n->children[n->numchildren-1]->numnodes;

    } else if (c == ')') {	/* End of node, null		*/
      
      /* Shrink children to necessary number, if any exist	*/
      /* If not, it's a leaf, so add itself to its subtree	*/

      if (n->children != 0) {
	(n->numnodes)++;	/* This node itself is no leaf	*/
	n->children = (nodeptr*)realloc( (void*) n->children,
					 n->numchildren * sizeof(nodeptr));
      } else {
	n->numleafs = 1;
	n->numnodes = 0;
      }
      n->procnum = procnum;	/* Remember procnum/cpu time if found	*/
      n->cputime = cputime;
      if (procnum > maxprocnum)	/* Update global maximum-proc.-index	*/
	  maxprocnum = procnum;

      return n;

    } else if ((c >= '0') && (c <= '9')) {

      /* Parse procedure number resp. CPU time */

      if (haveprocnum) {	/* Procnum read, now comes the CPU time */
	cputime = cputime * 10 + (long) (c - '0');
      } else {			/* Still in procnum			*/
	procnum = procnum * 10 + (int) (c - '0');
      }


    } else if (c == ':') {

      haveprocnum = 1;		/* End of procedure number seen		*/

    } else {

      printf("parse_tree: unexpected character '%c'\n", c);

    }

  }

  if (procnum > maxprocnum)	/* Update global maximum-proc.-index	*/
      maxprocnum = procnum;
  
  n->procnum = procnum;		/* Remember procnum/cpu time if found	*/
  n->cputime = cputime;
  return n;

}


/* Collects procedure-specific information (number of leafs/nodes
 * per procedure type, leaf = no other procedures of this type in subtree)
 */
void compute_proc_infos(nodeptr n) {
    int i;
    int t;
    
    /* Allocate & initialize space for procedure-specific statistics	*/

    if (maxprocnum == 0) {
	printf("warning - compute_proc_infos called but no procedure numbers given\n");
	return;
    }
    n->procnodes = (int*) malloc(maxprocnum * sizeof(int));
    n->procleafs = (int*) malloc(maxprocnum * sizeof(int));
    for (t=0; t<maxprocnum; t++) {
	n->procnodes[t] = 0;
	n->procleafs[t] = 0;
    }

    for (i=0; i<n->numchildren; i++) {	/* Get Infos recursively	*/
	compute_proc_infos(n->children[i]);
    }
    for (t=0; t<maxprocnum; t++) {	/* Compute #s for each proctype */
	for (i=0; i<n->numchildren; i++) {
	    n->procnodes[t] = n->procnodes[t] +
		              n->children[i]->procnodes[t];
	    n->procleafs[t] = n->procleafs[t] +
		              n->children[i]->procleafs[t];
	}
	if (n->procnum-1 == t) {	/* Regard the currend proc.type	*/
	    if (n->procleafs[t] == 0) {	/* No sub-leafs? Then this one	*/
		n->procleafs[t] = 1;	/*  is a "leaf" of this proctype*/
	    } else {
		(n->procnodes[t])++;	/* Else one more node		*/
	    }
	}
    }

}
    
/* Outputs leaf/node counts for each procedure, (12 chars for each proc)
 */
void print_proc_infos(nodeptr n) {
    int t;

    /* (here, t runs from 0 to maxprocnum-2, since maxprocnum is
     * always >0 if procnums are given at all, so index 0 means 1 etc)
     */
    for (t=0; t<maxprocnum; t++) {
	if ((n->procnodes) && (n->procleafs)) {		/* (sanity check) */
	    printf("%c|%4d/%-4d ", (char) ('A'+t),
		   n->procnodes[t], n->procleafs[t]);
	} else {
	    printf("(procinfos?)");
	}
    }
}


/* Outputs the tree semi-graphically with the number of
 * leafs and nodes in the subtree. If verbose is true,
 * also output procedure number (as "A"..."Z") and CPU time used.
 * Setting procverbose to true results in subtree-node/leaf output
 * for each recursive procedure type.
 */
void print_tree(nodeptr n, int depth, char first,
		char verbose, char procverbose)
{
    int t;
    int i;
    int j;
    char any;
    char f;

    if (!first)
      for (t=0; t< depth; t++) {        /* Indent for depth     	*/
	  if ((verbose==1) && (procverbose==0)) {
	      printf("                   ");
	} else if ((verbose==0) && (procverbose==1)) {
	  printf("            ");
	  for (j=0; j<maxprocnum; j++)	/* Room for procverbose		*/
	      printf("            ");
	} else if ((verbose==1) && (procverbose==1)) {
	  for (j=0; j<maxprocnum; j++)	/* Room for procverbose		*/
	      printf("            ");
	} else {
	  printf("            ");
	}
      }
    
    if (verbose && !procverbose) {
	printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
	       n->cputime, n->numnodes, n->numleafs);
    } else if (!verbose && procverbose) {
	printf(" %5d/%-5d", n->numnodes, n->numleafs);
	print_proc_infos(n);
    } else if (verbose && procverbose) {
	printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
	       n->cputime, n->numnodes, n->numleafs);
	print_proc_infos(n);
    } else { 
	printf(" %5d/%-5d", n->numnodes, n->numleafs);
    }
    f=1;

    for (i=0; i<n->numchildren; i++) {
      print_tree(n->children[i], depth+1, f, verbose, procverbose);
      f=0; 	                   	/* Reset flag after 1st print	*/
    }
    if (!n->numchildren)
      printf("\n");             	/* New line if no children  	*/
    
}


/* Print tree graphically using ASCII output.
 * If verbose flag is set, print procedure number (A,B,...)
 * instead of node symbol.
 */
void print_tree_graphics(nodeptr n, int depth,
			 char first, char last, char* branches,
			 char verbose)
{
    int t;
    int i;
    char f;	/* Flag: True if first child of this node	*/
    char l;	/* Flag: True if last child of this node	*/
    char any;

    if (!first) {
      for (t=0; t < depth; t++)		/* Indent for depth	*/
	  printf("%c  ", branches[t]);
    }
    
    f=1;


    for (i=0; i<n->numchildren; i++) {	/* Lengthen branch if any subbranches*/
	l = (char)(i >= n->numchildren-1);
	if (f) {
	    if (verbose)		/* Procedure number as character*/
	        printf("%c--", (char)(n->procnum + 'A'-1));
	    else
	        if ((last) && (!first))
		    printf("`--");	/* Last branch looks different	*/
	        else	
		    printf("+--");
	    branches[depth+1]='|';
	}				/* Set flag for last child	*/
	if (l)				/* Only reset after last child	*/
	    branches[depth+1]=' ';
	print_tree_graphics(n->children[i], depth+1,
			    f, l, branches, verbose);
	f=0;
    }
    if (!n->numchildren)		/* Print leaf correctly		*/
        if (verbose) {
	    printf("%c\n", (char)(n->procnum + 'A'-1));
	} else {
	    if ((last) && (!first))
	        printf("`-O\n"); 
	    else
	        printf("+-O\n"); 
	}

}


/* Compute depth of tree recursively
 * (a one-leaf-only tree has depth 1 but its node(leaf) will be at
 * level 0, i.e. the depths of the tree range from 0..tree_depth(root) !)
 */
int tree_depth(nodeptr n)
{
    int i;
    int d;
    int cd;

    d=1;				/* Own depth-increase		*/
    for (i=0; i<n->numchildren; i++) {
	cd = tree_depth(n->children[i]);
	if (cd+1 > d)			/* Remember maximum child depth	*/
	    d=cd+1;
    }
    return d;
   
}


/* Compute maximum branching degree of tree recursively
 */
int tree_degree(nodeptr n)
{
    int i;
    int d;
    int cd;

    d=n->numchildren;			/* Own degree			*/
    for (i=0; i<n->numchildren; i++) {
	cd = tree_degree(n->children[i]);
	if (cd > d)			/* Remember maximum child degree*/
	    d=cd;
    }
    return d;
   
}


/* Prints statistics on number of nodes/leafs in subtrees
 * at given depth immediately
 */
void print_subtrees_at_depth(nodeptr n, int depth,
			     char verbose, char procverbose)
{
    if (depth<=0) {
	if (verbose && !procverbose) {
	    printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
		   n->cputime, n->numnodes, n->numleafs);
	} else if (!verbose && procverbose) {
	    printf(" %5d/%-5d", n->numnodes, n->numleafs);
	    print_proc_infos(n);
	} else if (verbose && procverbose) {
	    printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
		   n->cputime, n->numnodes, n->numleafs);
	    print_proc_infos(n);
	} else { 
	    printf(" %5d/%-5d", n->numnodes, n->numleafs);
	}
    } else {
	int i;
	for (i=0; i<n->numchildren; i++)
	    print_subtrees_at_depth(n->children[i], depth-1,
				    verbose, procverbose);
    }
}


/* Collects statistics on number of nodes/leafs in subtrees
 * at given depth and returns pseudo-node with array of pointers to nodes.
 * info[1] gives number of nodes at exactly this depth,
 * info[2] ditto for leafs (info[0] is used internally)
 */
nodeptr subtrees_at_depth(nodeptr n, int depth)
{
    nodeptr statnode;
    int     t;
    
    statnode = (nodeptr) malloc(sizeof(node));
    statnode->numchildren = 0;		/* Initialize values		*/
    statnode->children = (nodeptr*)0;
    statnode->numnodes = 0;
    statnode->numleafs = 0;
    statnode->info     = (int*)malloc(3*sizeof(int));
    statnode->info[0]  = 0;	/* # of currently allocated children	*/
    statnode->info[1]  = 0;	/* # of nodes at exactly this depth	*/
    statnode->info[2]  = 0;	/* # of leafs ditto			*/
    statnode->procleafs = (int*) malloc(maxprocnum * sizeof(int));
    statnode->procnodes = (int*) malloc(maxprocnum * sizeof(int));

    for (t=0; t<maxprocnum; t++) {
	statnode->procnodes[t] = 0;
	statnode->procleafs[t] = 0;
    }

    get_subtrees_at_depth(n, depth, statnode);

    /* Shrink array of children to necessary size		*/
    
    if (statnode->children != 0) {
	statnode->children = (nodeptr*)realloc( (void*) statnode->children,
						statnode->numchildren *
						sizeof(nodeptr));
    }

    return statnode;
}


void get_subtrees_at_depth(nodeptr n, int depth, nodeptr statnode)
{
    int t;
    
    if (depth < 0)
	return;
    if (depth == 0) {		/* Depth found, add stats to node*/

      (statnode->numchildren)++;
      if (statnode->children == 0) {
	
	  /* Didn't have children yet, allocate some space	*/

	  statnode->children = (nodeptr*)malloc(CHILDNUM * sizeof(nodeptr));
	  statnode->info[0] = CHILDNUM;/* Remember current max.alloc.*/

      } else if (statnode->numchildren >= statnode->info[0]) {
	
	/* More children than already allocated, make room	*/

	statnode->info[0] = statnode->info[0] + CHILDNUM;
	statnode->children = (nodeptr*)realloc( (void*) statnode->children,
						statnode->info[0] *
						sizeof(nodeptr));
      }

      statnode->children[statnode->numchildren-1] = n;
      statnode->numleafs = statnode->numleafs +		/* Count leafs	*/
	                   n->numleafs;
      statnode->numnodes = statnode->numnodes +		/* Count nodes	*/
  	                   n->numnodes;
      for (t=0; t<maxprocnum; t++) {	/* Compute #s for each proctype */
	  statnode->procnodes[t] = statnode->procnodes[t] +
	                           n->procnodes[t];
	  statnode->procleafs[t] = statnode->procleafs[t] +
		                   n->procleafs[t];
      }
      if (n->numchildren == 0)
	  (statnode->info[2])++;      /* Count leafs at THIS depth only */
      else
	  (statnode->info[1])++;      /* Ditto nodes			*/

    } else {
	int i;
	for (i=0; i<n->numchildren; i++)
	    get_subtrees_at_depth(n->children[i], depth-1, statnode);
    }
}


/* Collects statistics on number of nodes/leafs in subtrees
 * of the given procedure type (procnum) whose depth is less than or
 * equal the given depth returns pseudo-node with array of pointers to nodes.
 * info[1] gives number of nodes of given type until this depth,
 * info[2] ditto for leafs (info[0] is used internally)
 */
nodeptr procnum_subtrees_till_depth(nodeptr n, int procnum, int depth)
{
    nodeptr statnode;
    int     t;
    
    statnode = (nodeptr) malloc(sizeof(node));
    statnode->numchildren = 0;		/* Initialize values		*/
    statnode->children = (nodeptr*)0;
    statnode->numnodes = 0;
    statnode->numleafs = 0;
    statnode->info     = (int*)malloc(3*sizeof(int));
    statnode->info[0]  = 0;	/* # of currently allocated children	*/
    statnode->info[1]  = 0;	/* # of nodes until this depth		*/
    statnode->info[2]  = 0;	/* # of leafs ditto			*/
    statnode->procleafs = (int*) malloc(maxprocnum * sizeof(int));
    statnode->procnodes = (int*) malloc(maxprocnum * sizeof(int));

    for (t=0; t<maxprocnum; t++) {
	statnode->procnodes[t] = 0;
	statnode->procleafs[t] = 0;
    }

    get_procnum_subtrees_till_depth(n, procnum, depth, statnode);

    /* Shrink array of children to necessary size		*/
    
    if (statnode->children != 0) {
	statnode->children = (nodeptr*)realloc( (void*) statnode->children,
						statnode->numchildren *
						sizeof(nodeptr));
    }

    return statnode;
}


void get_procnum_subtrees_till_depth(nodeptr n, int procnum, int depth,
				     nodeptr statnode)
{
    int i,t;
    
    if (depth < 0)
	return;
    
    /* Depth still OK and right type of node? add stats to statnode! */
    
    if ((depth >= 0) && (n->procnum == procnum)) {

      (statnode->numchildren)++;
      if (statnode->children == 0) {
	
	  /* Didn't have children yet, allocate some space	*/

	  statnode->children = (nodeptr*)malloc(CHILDNUM * sizeof(nodeptr));
	  statnode->info[0] = CHILDNUM;/* Remember current max.alloc.*/

      } else if (statnode->numchildren >= statnode->info[0]) {
	
	/* More children than already allocated, make room	*/

	statnode->info[0] = statnode->info[0] + CHILDNUM;
	statnode->children = (nodeptr*)realloc( (void*) statnode->children,
						statnode->info[0] *
						sizeof(nodeptr));
      }

      statnode->children[statnode->numchildren-1] = n;
      statnode->numleafs = statnode->numleafs +		/* Count leafs	*/
	                   n->numleafs;
      statnode->numnodes = statnode->numnodes +		/* Count nodes	*/
  	                   n->numnodes;
      for (t=0; t<maxprocnum; t++) {	/* Compute #s for each proctype */
	  statnode->procnodes[t] = statnode->procnodes[t] +
	                           n->procnodes[t];
	  statnode->procleafs[t] = statnode->procleafs[t] +
		                   n->procleafs[t];
      }
      if (n->numchildren == 0) {
	  (statnode->info[2])++;      /* Count leafs until this depth	*/
      } else {
	  (statnode->info[1])++;      /* Ditto nodes			*/
      }
    }
    
    /* Check all children, too (depth already checked above) 	*/
    
    for (i=0; i<n->numchildren; i++)
	get_procnum_subtrees_till_depth(n->children[i], procnum,
					depth-1, statnode);
    
}


/* Compute average of integer array
 */
double avg(int *array, int n)
{
    int    i;
    double a;

    a=0.0;

    for (i=0; i<n; i++)
	a = a + (double)array[i];

    return a/(double)n;
    
}


/* Given the array and the average, computes the deviation
 */
double dev(int *array, int n, double a)
{
    int    i;
    double d;

    d=0.0;

    for (i=0; i<n; i++)
	d = d + ((double)array[i]-a)*((double)array[i]-a);

    return sqrt(d/(double)n);

}


main(int argc, char **argv)
{
  char    tree_out       = 0;	/* Output tree?				*/
  char	  tree_verbose   = 0;	/* Verbose tree output (timings etc) ?	*/
  char	  grtree_out     = 0;	/* Ditto graphical tree			*/
  char	  grtree_verbose = 0;
  char    stats_out      = 0;	/* Ditto statistics			*/
  char	  stats_verbose  = 0;
  char    debug_out      = 0;	/* More detailed output			*/
  
  nodeptr root;
  nodeptr *subroot;
  char   *inputtext = (char*) malloc(1);
  char    inputbuf[TEXTCHUNKSIZE];
  int     inputlen = 0;
  int     l;
  int     i,j;
  int     depth;
  nodeptr n;
  int	 *depthnodes;
  int    *depthleafs;
  char    branchtext[BUFSIZ];
  int   **leafsizes;		/* Arrays for # of leafs/nodes		*/
  int   **nodesizes;		/*  in subtrees at each depth		*/
  int    *treenum;		/* Number of subtrees for each depth	*/
  int    *allleafsizes;		/* Arrays for # of ALL leafs/nodes	*/
  int    *allnodesizes;
  int     allnodenum;
  int     allleafs;
  int     allnodes;
  int     minnodes;
  int     mednodes;		/* Minimum, median, maximum	*/
  int     maxnodes;
  double  avgnodes;		/* Average, Std.deviation	*/
  double  devnodes;
  int     minleafs;
  int     medleafs;
  int     maxleafs;
  double  avgleafs;
  double  devleafs;

  int     procnum = 4;		/* Number of processors for parallelization */
  double  procpercent = 100.0 / (double) procnum;/* Percentage of 1 CPU      */


  for (i=1; i<argc; i++) {
      if (!strcmp(argv[i],"-t")) {
	  printf("Tree output\n");
	  tree_out = 1;
      }
      if (!strcmp(argv[i],"-tv")) {
	  printf("Tree output, verbose\n");
	  tree_out = 1;
	  tree_verbose = 1;
      }
      if (!strcmp(argv[i],"-g")) {
	  printf("Graphical tree output\n");
	  grtree_out = 1;
      }
      if (!strcmp(argv[i],"-gv")) {
	  printf("Graphical tree output, verbose\n");
	  grtree_out = 1;
	  grtree_verbose = 1;
      }
      if (!strcmp(argv[i],"-s")) {
	  printf("Statistics output\n");
	  stats_out = 1;
      }
      if (!strcmp(argv[i],"-sv")) {
	  printf("Statistics output, verbose\n");
	  stats_out = 1;
	  stats_verbose = 1;
      }
      if (!strcmp(argv[i],"-d")) {
	  printf("Detailed output\n");
	  debug_out = 1;
      }
      if ( (!strcmp(argv[i],"-h")) || (!strcmp(argv[i],"-help")) ) {
	  printf("%s - parse recorded recursion tree from stdin,\n", argv[0]);
	  printf("optionally with recorded procedure numbers and timings\n\n");
	  printf("Program options:\n");
	  printf("  -d  output detailed information\n");
	  printf("  -g  output graphical tree\n");
	  printf("  -gv ditto verbose (procedure name tags)\n");
	  printf("  -s  output statistics\n");
	  printf("  -sv ditto verbose (sizes of subtrees)\n");
	  printf("  -t  output tree\n");
	  printf("  -tv ditto verbose (timings etc)\n");
	  printf("\n");
	  exit(123);
      }
  }
      
  *inputtext = '\0';
  while (feof(stdin) == 0) {
      if (fgets(inputbuf, TEXTCHUNKSIZE, stdin)) {
	  l = strlen(inputbuf);
	  inputlen = inputlen + l;
	  inputtext = (char*)realloc(inputtext, inputlen);
/*	  printf("read %d bytes, chunk size = %d\n", l, TEXTCHUNKSIZE);*/
	  strcat(inputtext, inputbuf);
      }
  }
  if (strlen(inputtext)>1)
      text = inputtext+1;	/* Skip first dummy-root-node		*/
  else
      text = inputtext;

  maxprocnum = 0;		/* Init maximum procedure number	*/
  root = parse_tree();
  compute_proc_infos(root);	/* Get subtree nodes/leafs for each proc*/
  
  if (debug_out) {
      printf("Tree as text = '%s'\n",inputtext);
  }
  if (tree_out) {
      if (tree_verbose) {
	  printf("\nTree indented = (proc: time nodes/leafs)\n");
	  print_tree(root, 0, 1, 1, 1);
      } else {
	  printf("\nTree indented = (nodes/leafs)\n");
	  print_tree(root, 0, 1, 0, 0);
      }
  }
  if (grtree_out) {
      for (i=0; i<BUFSIZ; i++)
	  branchtext[i] = ' ';
      if (grtree_verbose) {
	  printf("\nTree graphically (procedures as characters) =\n");
	  print_tree_graphics(root, 1, 0, 1, branchtext, 1);
      } else {
	  printf("\nTree graphically =\n");
	  print_tree_graphics(root, 1, 0, 1, branchtext, 0);
      }
  }
  
  depth=tree_depth(root);
  i=tree_degree(root);
  if (stats_out) {
      printf("\nTree depth  = %d\n", depth);
      printf("Tree degree = %d\n", i);
      if (root->cputime > 0) {
	  printf("ms per node = %-12.6f\n",
		 (double)root->cputime / (double)root->numnodes);
	  printf("ms per leaf = %-12.6f\n",
		 (double)root->cputime / (double)root->numleafs);
      } else {
	  printf("no timings given\n");
      }
      printf("Maximum of %d different procedures in tree\n", maxprocnum);
  }

  allnodenum = 0;
  allnodes   = 0;
  allleafs   = 0;

  nodesizes = (int**) malloc(depth * sizeof(int*));
  leafsizes = (int**) malloc(depth * sizeof(int*));
  treenum   = (int*)  malloc(depth * sizeof(int));
  subroot   = (nodeptr*) malloc (depth * sizeof(nodeptr));
  
  for (i=1; i<depth; i++) {
      int    maxn;
      
      n = subtrees_at_depth(root, i);
      subroot[i] = n;
      treenum[i] = n->numchildren;

      /* Fill arrays with numbers of nodes/leafs in subtree,
       * compute averages and standard deviation of node/leaf numbers
       */
      nodesizes[i] = (int*) malloc (treenum[i] * sizeof(int));
      leafsizes[i] = (int*) malloc (treenum[i] * sizeof(int));
      for (j=0; j<treenum[i]; j++) {
	  nodesizes[i][j] = n->children[j]->numnodes;
	  leafsizes[i][j] = n->children[j]->numleafs;
      }
      avgnodes = avg(nodesizes[i], treenum[i]);
      avgleafs = avg(leafsizes[i], treenum[i]);
      devnodes = dev(nodesizes[i], treenum[i], avgnodes);
      devleafs = dev(leafsizes[i], treenum[i], avgleafs);
      
      /* Sort the arrays, then easily read the minimum,
       * median and maximum number of subnodes/leafs
       */
      intsort(nodesizes[i], treenum[i]);
      intsort(leafsizes[i], treenum[i]);

      minnodes = nodesizes[i][0];
      maxnodes = nodesizes[i][treenum[i]-1];
      mednodes = nodesizes[i][treenum[i]/2];
      minleafs = leafsizes[i][0];
      maxleafs = leafsizes[i][treenum[i]-1];
      medleafs = leafsizes[i][treenum[i]/2];

      allnodes = allnodes + n->info[1];
      allleafs = allleafs + n->info[2];

      /* Output the collected statistics
       * and the first few subtree summary nodes
       */
      if (stats_out) {
	  printf("\n");
	  printf("Depth %2d: %5d nodes and %5d leafs in %5d subtrees\n",
		 i, n->numnodes, n->numleafs, treenum[i]);
	  printf("          %5d nodes and %5d leafs at exactly this depth\n",
		 n->info[1], n->info[2]);
	  printf(" subnodes: %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
		 minnodes, mednodes, maxnodes, avgnodes, devnodes);
	  if (avgnodes > 0)
	      printf(" (%6.3f)\n", devnodes/avgnodes);
	  else
	      printf("\n");
	  printf(" subleafs: %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
		 minleafs, medleafs, maxleafs, avgleafs, devleafs);
	  if (avgleafs > 0)
	      printf(" (%6.3f)\n", devleafs/avgleafs);
	  else
	      printf("\n");

	  if (stats_verbose) {
	      if (treenum[i] > 16) {
		  printf("(showing just the first 16)\n");
		  maxn=16;
	      } else
		  maxn = treenum[i];
      
	      for (j=0; j<maxn; j++) {
		  printf(" %c:%5dms %5d/%-5d",
			 (char) ('A'-1+n->children[j]->procnum),
			 n->children[j]->cputime,
			 n->children[j]->numnodes, n->children[j]->numleafs);
		  print_proc_infos(n->children[j]);
		  printf("\n");
	      }
	  }
      }

      /* Keep all-node/leaf-information for final statistics
       */
      if (allnodenum)  {
	  allnodesizes = (int*) realloc(allnodesizes,
					(allnodenum + treenum[i]) *
					sizeof(int));
	  allleafsizes = (int*) realloc(allleafsizes,
					(allnodenum + treenum[i]) *
					sizeof(int));
      } else {
	  allnodesizes = (int*) malloc(treenum[i] * sizeof(int));
	  allleafsizes = (int*) malloc(treenum[i] * sizeof(int));
      }
      for (j=0; j<treenum[i]; j++) {
	  allnodesizes[allnodenum+j] = nodesizes[i][j];
	  allleafsizes[allnodenum+j] = leafsizes[i][j];
      }
      allnodenum = allnodenum + treenum[i];

  }

  /* Now take a look at all procedure types and their subtrees
   * at any depth
   */
  for (i=1; i<=maxprocnum; i++) {
      
      n = procnum_subtrees_till_depth(root, i, depth);

      if (stats_out) {
	  int maxn;
	  
	  printf("\n");
	  printf("Procedure %c: %5d nodes and %5d leafs in %5d subtrees\n",
		 (char) ('A'-1+i), n->numnodes, n->numleafs, n->numchildren);
	  printf("             %5d nodes and %5d leafs of this procedure until this depth\n",
		 n->info[1], n->info[2]);

	  if (stats_verbose) {
	      if (n->numchildren > 16) {
		  printf("(showing just the first 16)\n");
		  maxn=16;
	      } else
		  maxn = n->numchildren;
      
	      for (j=0; j<maxn; j++) {
		  printf(" %c:%5dms %5d/%-5d",
			 (char) ('A'-1+n->children[j]->procnum),
			 n->children[j]->cputime,
			 n->children[j]->numnodes, n->children[j]->numleafs);
		  print_proc_infos(n->children[j]);
		  printf("\n");
	      }
	  }
      }

      if (n->children)
	  free(n->children);
      free(n);
  }
  
  /* Sort the all-nodes/leafs arrays arrays, then easily read the minimum,
   * median and maximum number of subnodes/leafs
   */
  intsort(allnodesizes, allnodenum);
  intsort(allleafsizes, allnodenum);
  
  minnodes = allnodesizes[0];
  maxnodes = allnodesizes[allnodenum-1];
  mednodes = allnodesizes[allnodenum/2];
  minleafs = allleafsizes[0];
  maxleafs = allleafsizes[allnodenum-1];
  medleafs = allleafsizes[allnodenum/2];

  avgnodes = avg(allnodesizes, allnodenum);
  avgleafs = avg(allleafsizes, allnodenum);
  devnodes = dev(allnodesizes, allnodenum, avgnodes);
  devleafs = dev(allleafsizes, allnodenum, avgleafs);

  printf("\nComplete Tree (all %d subtrees together, %d nodes, %d leafs):\n",
	 allnodenum, allnodes, allleafs);
  printf(" nodes:    %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
	 minnodes, mednodes, maxnodes, avgnodes, devnodes);
  if (avgnodes > 0)
      printf(" (%6.3f)\n", devnodes/avgnodes);
  else
      printf("\n");
  printf(" leafs:    %5d min, %5d med, %5d max, %8.2f avg, %8.3f dev",
	 minleafs, medleafs, maxleafs, avgleafs, devleafs);
  if (avgleafs > 0)
      printf(" (%6.3f)\n", devleafs/avgleafs);
  else
      printf("\n");

  /* Heuristic for Depth choice: Take first depth (from root)
   * where
   *	- Number of subtrees >= number of CPUs
   *	- Percentage of nodes in largest subtree relative to all subtrees
   *	   <= 1/number of CPUs  (to make sure work is distributed equally)
   *	- Ditto for leafs
   */
  printf("\nRecommended depth for DEPTH_N strategy on %d processors (%5.2f%% per CPU):\n",
	 procnum, procpercent);
  for (i=1; i<depth; i++) {
      double subnodepercent;	/* Percentage of # subnodes of largest	*/
      double subleafpercent;	/*  node to # of all subnodes, & leafs	*/
      double treenodepercent;	/* Percentage of nodes in this subtree	*/
      double treeleafpercent;	/*  to all nodes in the whole tree/leafs*/
      int    ok = 0;

      printf("\nDepth %2d : ", i);
      printf("Number of subtrees = %5d\n", treenum[i]);

      treenodepercent = 100.0 * ((double)subroot[i]->numnodes) /
	                        ((double) root->numnodes);
      treeleafpercent = 100.0 * ((double)subroot[i]->numleafs) /
	                        ((double) root->numleafs);
      printf("  Nodes at and below this depth = %6.2f%% of all nodes\n",
	     treenodepercent);
      printf("  Leafs at and below this depth = %6.2f%% of all leafs\n",
	     treeleafpercent);


      /* subtree with max. nodes does not have to be the same as the
       * one with max. leafs!! SO WHAT?
       */
      if (subroot[i]->numnodes) {
	  subnodepercent = 100.0 * (double)nodesizes[i][treenum[i]-1] /
	                           (double) subroot[i]->numnodes;
      } else {
	  subnodepercent = 0.0;
      }
      if (subroot[i]->numleafs) {
	  subleafpercent = 100.0 * (double)leafsizes[i][treenum[i]-1] /
	                           (double) subroot[i]->numleafs;
      } else {
	  subleafpercent = 0.0;
      }
      printf("  Largest subtree has %5d nodes (%6.2f%%) ",
	     nodesizes[i][treenum[i]-1], subnodepercent);
      printf("  and %5d leafs (%6.2f%%)\n",
	     leafsizes[i][treenum[i]-1], subleafpercent);

      if (treenum[i] >= procnum) {
	  printf("  OK : ");
	  ok++;
      } else {
	  printf("  not: ");
      }
      printf("# subtrees >= # processors\n");
      if (treenodepercent >= procpercent) {
	  printf("  OK : ");
	  ok++;
      } else {
	  printf("  not: ");
      }
      printf("subtree node percentage >= 1 CPU percentage\n");
      if (treeleafpercent >= procpercent) {
	  printf("  OK : ");
	  ok++;
      } else {
	  printf("  not: ");
      }
      printf("subtree leaf percentage >= 1 CPU percentage\n");
      if (subnodepercent <= procpercent) {
	  printf("  OK : ");
	  ok++;
      } else {
	  printf("  not: ");
      }
      printf("node percentage of largest <= 1 CPU percentage\n");
      if (subleafpercent <= procpercent) {
	  printf("  OK : ");
	  ok++;
      } else {
	  printf("  not: ");
      }
      printf("leaf percentage of largest <= 1 CPU percentage\n");
      if (ok >= 5) {
	  printf("  -> RECOMMENDED\n");
      }
  }
  
}

